#ifndef _VEDIRECT_H_
#define _VEDIRECT_H_

#include <velib/velib_config_app.h>
#include <velib/ve_regs.h>
#include <velib/types/variant.h>
#include <velib/types/types.h>
#include <velib/vecan/products.h>
#include "vehex.h"
#include "vetext_fh.h"

#define SERIALNUMBER_LENGTH    20

typedef enum { VdUnkown, VdConnected, VdIdentified, VdDisConnected } VdDirectStatus;
typedef enum { VdHexTimeout, VdHexChecksum, VdHexBusy, VdHexRxBufferOverflow, VdHexTxBufferOverflow, VdHexSetTypeUnsupported } VdHexError;

typedef void (*VdFrameNotify)(veBool valid, void * context);
typedef void (*VdStatusChange)(VdDirectStatus status, void * context);
typedef void (*VdTextItemNotify)(VeRegId id, const VeVariantValue * value, VeDataBasicType type, void * context);
typedef void (*VdHexMessage)(VeRegId id, const un8 * data, un8 length, un8 flags, veBool async, void * context);
typedef void (*VdTxData)(const un8 * data, un8 length, void * context);
typedef void (*VdHexErrorNotify)(const VdHexError error, void * context);
typedef un16 (*VdRequestNotify)(CommandChars request, void * context);

typedef struct {
	VdFrameNotify     frameNotify;
	VdStatusChange    statusChange;
	VdTextItemNotify  textItemNotify;
	VdHexMessage      hexMessageNotify;
	VdTxData          txData;
	VdHexErrorNotify  hexErrorNotify;
	VdRequestNotify   hexRequestNotify;
} VdCallbackFunctions;

typedef enum {
	StatusChange, FrameNotify, TextItemNotify, TxData, HexErrorNotify, HexRequestNotify, HexMessageNotify
} VdCallbackIdentifiers;

typedef struct {
	void *context;
	un8 validFrameTimer;
	un8 hexTimer;
	VeRegId hexId;
	VdDirectStatus status;
	un16 prodId;
	enum { IdNone, IdVS, IdCE, IdPROD, IdBMV, IdPID } idPrecendence;
	un16 version;
	veBool hexBusy;
} VeDirectDevice;

typedef struct {
	VeDirectDevice device;
	VdCallbackFunctions callbacks;
	VdTextFhInstance fhInstance;
	VdTextInstance textInstance;
	VdHexInstance hexInstance;
} VeDirectInstance;

#ifdef __cplusplus
extern "C" {
#endif

/*
 * API Functions
 */

/* VE.Direct */
void   vdInit(VeDirectInstance * instance, void * context);
static inline void   vdSetContext(VeDirectInstance * instance, void * context) {instance->device.context = context;}
void   vdConnect(VeDirectInstance * instance, VdCallbackIdentifiers ident, void const * callback);
void   vdConnectAll(VeDirectInstance * instance, VdCallbackFunctions const * callbackFunctions);
void   vdDisconnect(VeDirectInstance * instance, VdCallbackIdentifiers ident);
void   vdDisconnectAll(VeDirectInstance * instance);
static inline veBool vdIsConnected(VeDirectInstance * instance) { return instance->device.status == VdConnected || instance->device.status == VdIdentified; }
static inline veBool vdIsIdentified(VeDirectInstance * instance)  { return instance->device.status == VdIdentified; }
void   vdSetValue(VeDirectInstance * instance, VeRegId id, const VeVariantValue * value, VeDataBasicType type);
void   vdSetData(VeDirectInstance * instance, VeRegId id, const un8 * data, un8 length);
void   vdGetValue(VeDirectInstance * instance, const VeRegId id);
static inline un16 vdGetProdId(VeDirectInstance * instance) { return instance->device.prodId; }
static inline veBool vdHexBusy(VeDirectInstance * instance) { return instance->device.hexBusy; }
void   vdTick(VeDirectInstance * instance);
static inline void vdRxData(VeDirectInstance * instance, un8 inbyte) { vhRxData(&instance->fhInstance, &instance->textInstance, &instance->hexInstance, inbyte); }
/* vetext functions */
veBool vdTextItemUpdate(VdTextInstance * instance, VeRegId id, const VeVariantValue * value, VeDataBasicType type);
void vdFrameReceived(VdTextInstance * instance, veBool valid);
VeVariantValue vdGetInvalidValue(VeDataBasicType type);

/* vehex functions */
void vdTxData(VdHexInstance * instance, const un8 * data, un8 length);
void vdHexErrorNotify(VdHexInstance * instance, VdHexError error);
un16 vdRequestNotify(VdHexInstance * instance, CommandChars request);
static inline void vdHexSendPing(VeDirectInstance * instance) { veHexSendPing(&instance->hexInstance);}
static inline void vdHexRestart(VeDirectInstance * instance) { veHexRestart(&instance->hexInstance);}
veBool vdHexItemUpdate(VdHexInstance * instance, HexMessage const * message);

#ifdef __cplusplus
}
#endif

#endif // _VEDIRECT_H_
