#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <velib/utils/ve_logger.h>
#include <velib/vecan/charger_error.h>
#include <velib/nmea2k/n2k.h>
#include "vedirect.h"
#include "vetext.h"
#include "prod_class.h"

#define MODULE "VE.Direct"

#define container_of(ptr, type, member) (type *)( (char *)ptr - offsetof(type,member) )

/*
 * Locals
 */

/*
 * Local prototype
 */
static const un8 frameTimerPreset = 50; /* 100 ms units -> timeout is 5 seconds */
static const un8 hexMessageTimeout = 10; /* 100 ms units -> timeout is 1 seconds */
static void veItemSetPidFromBMV(VeDirectInstance * instance, char const * value);
static void veItemSetPidFromProd(VeDirectInstance * instance, char const * value);

/*
 * Local functions
 */

static void veItemSetPidFromBMV(VeDirectInstance * instance, char const * const value)
{
	/* BMV-60x >= v2.08 have no PID field implemented, use the BMV field */
	if (strcmp(value, "600S") == 0)	{
		instance->device.prodId = VE_PROD_ID_BMV600S;
	} else if (strcmp(value, "602S") == 0) {
		instance->device.prodId = VE_PROD_ID_BMV602S;
	} else	if (strcmp(value, "600HS") == 0) {
		instance->device.prodId = VE_PROD_ID_BMV600HS;
	} else	if (strcmp(value, "700") == 0) {
		instance->device.prodId = VE_PROD_ID_BMV700;
	} else	if (strcmp(value, "700H") == 0) {
		instance->device.prodId = VE_PROD_ID_BMV700H;
	} else	if (strcmp(value, "702") == 0) {
		instance->device.prodId = VE_PROD_ID_BMV702;
	} else {
		logE(MODULE, "found unknown BMV");
	}
}

static void veItemSetPidFromProd(VeDirectInstance * instance, char const * const value)
{

	/* BlueSolar Chargers < v1.09 */
	if (strcmp(value, "MPPT 12..24V/15A") == 0) {
		instance->device.prodId = VE_PROD_ID_BLUE_SOLAR_MPPT_70_15;
		return;
	} else if (strcmp(value, "MPPT 12..24V/15A MK2") == 0) {
		instance->device.prodId = VE_PROD_ID_BLUE_SOLAR_MPPT_75_15;
		return;
	} else if (strcmp(value, "MPPT 12..24V/15A MK3") == 0) {
		instance->device.prodId = VE_PROD_ID_BLUE_SOLAR_MPPT_100_15;
		return;
	} else if (strcmp(value, "MPPT 12..24V/50A") == 0) {
		instance->device.prodId = VE_PROD_ID_BLUE_SOLAR_MPPT_75_50;
		return;
	} else {
		logE(MODULE, "found unknown BlueSolar Charger");
		return;
	}
}

static void vdSetStatus(VeDirectInstance * instance, VdDirectStatus status)
{
	instance->device.status = status;
	switch(status) {
	case VdDisConnected :
		veTextInit(&instance->textInstance);
		veHexInit(&instance->hexInstance);
		instance->device.idPrecendence = IdNone;
		instance->device.prodId = VE_PROD_NOT_SET;
		instance->device.hexBusy = veFalse;
		break;
	default:
		break;
	}
	if (instance->callbacks.statusChange) instance->callbacks.statusChange(status, instance->device.context);
}

/*
 * API for vedirect user
 */

void vdInit(VeDirectInstance * instance, void * const context)
{
	memset(instance, 0, sizeof(*instance));
	instance->device.context = context;
	instance->device.prodId = VE_PROD_NOT_SET;
	instance->device.idPrecendence = IdNone;
	instance->device.validFrameTimer = frameTimerPreset;
	instance->device.hexBusy = veFalse;
	vdSetStatus(instance, VdUnkown);
	vhInit(&instance->fhInstance);
	veTextInit(&instance->textInstance);
	veHexInit(&instance->hexInstance);
}

void vdConnect(VeDirectInstance * instance, const VdCallbackIdentifiers ident, void const * const callback)
{
	switch (ident) {
	case StatusChange:
		instance->callbacks.statusChange = (VdStatusChange) callback;
		break;
	case FrameNotify:
		instance->callbacks.frameNotify = (VdFrameNotify) callback;
		break;
	case TextItemNotify:
		instance->callbacks.textItemNotify = (VdTextItemNotify) callback;
		break;
	case TxData:
		instance->callbacks.txData = (VdTxData) callback;
		break;
	case HexErrorNotify:
		instance->callbacks.hexErrorNotify = (VdHexErrorNotify) callback;
		break;
	case HexRequestNotify:
		instance->callbacks.hexRequestNotify = (VdRequestNotify) callback;
		break;
	case HexMessageNotify:
		instance->callbacks.hexMessageNotify = (VdHexMessage) callback;
		break;
	}
}

void vdConnectAll(VeDirectInstance * instance, VdCallbackFunctions const * const callbackFunctions)
{
	instance->callbacks = *callbackFunctions;
}

void vdDisconnect(VeDirectInstance * instance, const VdCallbackIdentifiers ident)
{
	switch (ident) {
	case StatusChange:
		instance->callbacks.statusChange = NULL;
		break;
	case FrameNotify:
		instance->callbacks.frameNotify = NULL;
		break;
	case TextItemNotify:
		instance->callbacks.textItemNotify = NULL;
		break;
	case TxData:
		instance->callbacks.txData = NULL;
		break;
	case HexErrorNotify:
		instance->callbacks.hexErrorNotify = NULL;
		break;
	case HexRequestNotify:
		instance->callbacks.hexRequestNotify = NULL;
		break;
	case HexMessageNotify:
		instance->callbacks.hexMessageNotify = NULL;
		break;
	}
}

void vdDisconnectAll(VeDirectInstance * instance)
{
	memset(&instance->callbacks, 0, sizeof(instance->callbacks));
}

void vdSetValue(VeDirectInstance * instance, const VeRegId id, const VeVariantValue * value, const VeDataBasicType type)
{
	logI(MODULE, "Item changed %4.4X", id);

	if (instance->device.hexBusy) {
		vdHexErrorNotify(&instance->hexInstance, VdHexBusy);
		return;
	}
	instance->device.hexTimer = hexMessageTimeout;
	instance->device.hexId = id;
	switch (type) {
	case VE_SN8:
	case VE_UN8:
		instance->device.hexBusy = veTrue;
		veHexSet(&instance->hexInstance, id, (un8 const *)&value->UN8, sizeof(value->UN8));
		break;
	case VE_SN16:
	case VE_UN16:
		instance->device.hexBusy = veTrue;
		veHexSet(&instance->hexInstance, id, (un8 const *)&value->UN16, sizeof(value->UN16));
		break;
	case VE_SN32:
	case VE_UN32:
		instance->device.hexBusy = veTrue;
		veHexSet(&instance->hexInstance, id, (un8 const *)&value->UN32, sizeof(value->UN32));
		break;
	default:
		logE(MODULE, "Publish: Unsupported type %d", type);
		vdHexErrorNotify(&instance->hexInstance, VdHexSetTypeUnsupported);
		break;
	}
}

void vdSetData(VeDirectInstance * instance, const VeRegId id, const un8 * const data, const un8 length)
{
	if (instance->device.hexBusy) {
		vdHexErrorNotify(&instance->hexInstance, VdHexBusy);
		return;
	}
	instance->device.hexId = id;
	instance->device.hexBusy = veTrue;
	instance->device.hexTimer = hexMessageTimeout;
	veHexSet(&instance->hexInstance, id, data, length);
}

void vdGetValue(VeDirectInstance * instance, const VeRegId id)
{
	if (instance->device.hexBusy) {
		vdHexErrorNotify(&instance->hexInstance, VdHexBusy);
		return;
	}
	instance->device.hexId = id;
	instance->device.hexBusy = veTrue;
	instance->device.hexTimer = hexMessageTimeout;
	veHexGet(&instance->hexInstance, id);
}

void vdTick(VeDirectInstance * instance) /* 100 ms tick */
{
	if (instance->device.validFrameTimer)
		instance->device.validFrameTimer--;
	else {
		if (vdIsConnected(instance))
			vdSetStatus(instance, VdDisConnected);
	}
	if (instance->device.hexTimer)
		instance->device.hexTimer--;
	else {
		if (instance->device.hexBusy) {
			instance->device.hexBusy = veFalse;
			vdHexErrorNotify(&instance->hexInstance, VdHexTimeout);
		}
	}
}

/*
 * VE.Text functions
 */

/* Update item and remap or recalulate bugs or legacy stuff */
veBool vdTextItemUpdate(VdTextInstance * instance, const VeRegId id, const VeVariantValue * value, const VeDataBasicType type)
{
	VeDirectInstance * const vdInstance = container_of(instance, VeDirectInstance, textInstance);

	logI(MODULE,"RegId=%4.4X Value=%8.8X", id, value->UN32);

	un16 oldProdId = vdInstance->device.prodId;
	switch (id) {
	case VE_REG_PRODUCT_ID:
		if (vdInstance->device.idPrecendence <= IdPID) {
			vdInstance->device.idPrecendence = IdPID;
			vdInstance->device.prodId = value->UN16;
			if (oldProdId != vdInstance->device.prodId)
				vdSetStatus(vdInstance, VdIdentified);
		}
		break;
	case VE_REG_BMV:
		if (vdInstance->device.idPrecendence < IdBMV) {
			vdInstance->device.idPrecendence = IdBMV;
			veItemSetPidFromBMV(vdInstance, value->CPtr);
			if (oldProdId != vdInstance->device.prodId)
				vdSetStatus(vdInstance, VdIdentified);
		}
		/* Do not publish the BMV tag/reg. Only used to detemine product */
		return vdInstance->device.status != VdIdentified;
	case VE_REG_PROD:
		if (vdInstance->device.idPrecendence < IdPROD) {
			vdInstance->device.idPrecendence = IdPROD;
			veItemSetPidFromProd(vdInstance, value->CPtr);
			if (oldProdId != vdInstance->device.prodId)
				vdSetStatus(vdInstance, VdIdentified);
		}
		/* Do not publish the PROD tag/reg. Only used to detemine product */
		return vdInstance->device.status != VdIdentified;
	case VE_REG_DC_CHANNEL2_VOLTAGE:
		/* BMV-602 < v2.08 */
		if (vdInstance->device.idPrecendence < IdVS) {
			vdInstance->device.idPrecendence = IdVS;
			vdInstance->device.prodId = VE_PROD_ID_BMV602S;
			if (oldProdId != vdInstance->device.prodId)
				vdSetStatus(vdInstance, VdIdentified);
		}
		break;
	case VE_REG_CAH:
		/* BMV-600 < v2.08 */
		if (vdInstance->device.idPrecendence < IdCE) {
			vdInstance->device.idPrecendence = IdCE;
			vdInstance->device.prodId = VE_PROD_ID_BMV600S;
			if (oldProdId != vdInstance->device.prodId)
				vdSetStatus(vdInstance, VdIdentified);
		}
		break;
	case VE_REG_APP_VER:
		vdInstance->device.version = value->UN16;
		break;
	default:
		break;
	}

	/* Do not process item when not identified */
	if (vdInstance->device.status != VdIdentified)
		return veFalse;

	if (vdInstance->callbacks.textItemNotify) vdInstance->callbacks.textItemNotify(id, value, type, vdInstance->device.context);

	return veTrue;
}

void vdFrameReceived(VdTextInstance * instance, veBool valid)
{
	VeDirectInstance * const vdInstance = container_of(instance, VeDirectInstance, textInstance);

	if (valid) {
		vdInstance->device.validFrameTimer = frameTimerPreset;
		if (!vdIsConnected(vdInstance))
			vdSetStatus(vdInstance, VdConnected);
	}
	if (vdInstance->callbacks.frameNotify) vdInstance->callbacks.frameNotify(valid, vdInstance->device.context);
}

#if UINTPTR_MAX > UINT32_MAX
#define LARGE_FIELD Ptr
#else
#define LARGE_FIELD UN32
#endif

VeVariantValue vdGetInvalidValue(VeDataBasicType type)
{
	VeVariantValue value;
	value.LARGE_FIELD = 0;
	switch (type) {
	case VE_SN8:
		value.SN8 = VE_INVALID_SN8;
		break;
	case VE_UN8:
		value.UN8 = VE_INVALID_UN8;
		break;
	case VE_SN16:
		value.SN16 = VE_INVALID_SN16;
		break;
	case VE_UN16:
		value.UN16 = VE_INVALID_UN16;
		break;
	case VE_SN32:
		value.SN32 = VE_INVALID_SN32;
		break;
	case VE_UN32:
		value.UN32 = VE_INVALID_UN32;
		break;
	default:
		value.Ptr = NULL;
		break;
	}
	return value;
}

/*
 * VE.Hex functions
 */
void vdTxData(VdHexInstance * instance, const un8 * const data, const un8 length)
{
	const VeDirectInstance * const vdInstance = container_of(instance, VeDirectInstance, hexInstance);

	if (vdInstance->callbacks.txData) vdInstance->callbacks.txData(data,length,vdInstance->device.context);
}

void vdHexErrorNotify(VdHexInstance * instance, VdHexError error)
{
	VeDirectInstance * const vdInstance = container_of(instance, VeDirectInstance, hexInstance);

	if (vdInstance->callbacks.hexErrorNotify) vdInstance->callbacks.hexErrorNotify(error, vdInstance->device.context);
}

un16 vdRequestNotify(VdHexInstance * instance, CommandChars request)
{
	const VeDirectInstance * const vdInstance = container_of(instance, VeDirectInstance, hexInstance);
	if (vdInstance->callbacks.hexRequestNotify)
		return vdInstance->callbacks.hexRequestNotify(request, vdInstance->device.context);
	else
		return 0xFFFF;
}

veBool vdHexItemUpdate(VdHexInstance * instance, HexMessage const * const message)
{
	VeDirectInstance * const vdInstance = container_of(instance, VeDirectInstance, hexInstance);

	un8  * data = message->data;
	un8  length = message->length;
	const VeRegId id = message->veRegId;
	veBool async = message->response == CmdChar_NA_Notify;

	/* Preset frame timer when receiving hex messages to prevent text frame timeout */
	vdInstance->device.validFrameTimer = frameTimerPreset;

	if (message->response == CmdChar_Get_GetResponse
			|| message->response == CmdChar_Set_SetResponse)
	{
		if (vdInstance->device.hexId != id) {
			logE(MODULE, "Got a hex item update for a set or get of an id we did not request: %04X %04X", id, vdInstance->device.hexId);
		} else {
			vdInstance->device.hexBusy = veFalse;
		}
	}

	/* Do not process item when not identified */
	if (vdInstance->device.status != VdIdentified) return veFalse;

	switch (id) {
	case VE_REG_SERIAL_NUMBER:
		/* BMV sends 2 serial numbers: the first of the PCB and the second of the assembled
		 * product. The last one is also the one printed on the sticker and the one that
		 * should be reported. The two serial numbers are separated by a comma. */
		if ((length >= 23) && (data[11] == ',')) {
			length = 11;
			data += 12;
		} else if (length >= 11) {
			length = 11;
		}
		break;
	default:
		break;
	}

	/*
	 * The application should take care to copy the data if it should be memorized.
	 */
	if (vdInstance->callbacks.hexMessageNotify)
		vdInstance->callbacks.hexMessageNotify(id, data, length, message->flags, async, vdInstance->device.context);
	return veTrue;
}
