#include <string.h>

#include <velib/utils/ve_logger.h>

#include "vedirect_veti.h"
#include "vedirect.h"
#include "veti.h"

#define MODULE "vd2veti"

#if UINTPTR_MAX > UINT32_MAX
#define LARGE_FIELD Ptr
#else
#define LARGE_FIELD UN32
#endif

static inline const char * actionTypeStr(enum VdVetiActionType action) {
	return action == VdVetiGet ? "get" : "set";
}

static inline unsigned int actionIdx(unsigned int action, unsigned int offset) {
	return (action + offset) % CFG_VD_VETI_ACTIONS;
}

static inline void errorNotify(VdVetiInstance * const instance, const VdVetiError error) {
	if (instance->errorNotify)
		instance->errorNotify(error, instance->context);
}

static veBool signalActionsFull(VdVetiInstance * instance, enum VdVetiActionType newAction, const VetiId id)
{
	if (vdVetiActionsFull(instance)) {
#if CFG_WITH_LOG_PROTO
		const VdVetiAction * const curAction = &instance->actions[instance->actionsHead];
		logE(MODULE, "Trying to add a %s for %04X, but the action list is full: %d %s %04X",
			 actionTypeStr(newAction), id,
			 instance->actionsCount, actionTypeStr(curAction->actionType), curAction->id);
#endif
		errorNotify(instance, VdVetiActionListFull);

		return veTrue;
	} else {
		return veFalse;
	}
}

static void performAction(VdVetiInstance * const instance)
{
	const VdVetiAction * const action = &instance->actions[instance->actionsHead];

	logI(MODULE, "Performing action: %s %04X", actionTypeStr(action->actionType), action->id);

	if (action->actionType == VdVetiGet) {
		vdGetValue(instance->vdInstance, action->id);
	} else {
		vdSetValue(instance->vdInstance, action->id, &action->value, vetiToVeType(vetiGetType(instance->vetiInstance, action->id)));
	}
}

static void nextAction(VdVetiInstance * instance) {
	const VdVetiAction * const curAction = &instance->actions[instance->actionsHead];
	logI(MODULE, "Action done: %s %04X", actionTypeStr(curAction->actionType), curAction->id);

	instance->actionsHead = actionIdx(instance->actionsHead, 1);
	if (--instance->actionsCount) {
		performAction(instance);
	}
}

static void setToGet(VdVetiInstance * instance) {
	VdVetiAction * const action = &instance->actions[instance->actionsHead];
	logI(MODULE, "Turning set into get: %04X", action->id);
	action->actionType = VdVetiGet;
	performAction(instance);
}

VeProdClasses vdVetiClass(un16 productId)
{
	switch(productId) {
	case VE_PROD_ID_BMV600HS:
	case VE_PROD_ID_BMV600S:
	case VE_PROD_ID_BMV602S:
	case VE_PROD_ID_BMV700:
	case VE_PROD_ID_BMV700H:
	case VE_PROD_ID_BMV702:
		return classBattery;
	default:
		break;
	}

	/* Check for Smart BMV's */
	if ((productId & ~0x000F) == 0xA380)
		return classBattery;

	/* Check for MPPTs */
	if ((productId & ~0x003F) == 0xA040)
		return classSolarCharger;

	/* For now, use a range of 0x100 products */
	if ((productId >= VE_PROD_ID_PHOENIX_INVERTER)
		 && (productId < VE_PROD_ID_PHOENIX_INVERTER + 0x100))
		return classInverter;

	return classUndefined;
}

void vdVetiStatusChange(VdDirectStatus status, VdVetiInstance * instance)
{
	if (status == VdIdentified) {
		VeVariantValue v;
		v.UN16 = vdGetProdId(instance->vdInstance);
		vetiSetClass(instance->vetiInstance, vdVetiClass(v.UN16));
		vetiPublish(instance->vetiInstance, instance->veDirectVetiHandle, VE_REG_PRODUCT_ID, &v);
	} else {
		instance->actionsCount = 0;
	}
}

void vdVetiTextItemNotify (const VeRegId id, const VeVariantValue * value, const VeDataBasicType type, VdVetiInstance * instance)
{
	VE_UNUSED(type);
	if (vetiPublish(instance->vetiInstance, instance->veDirectVetiHandle, id, value) == veFalse)
		logW(MODULE, "RegId %4.4X not found!", id);
}

void vdVetiHexMessageNotify (const VeRegId id, const un8 * const data, const un8 length, const un8 flags, const veBool async, VdVetiInstance * const instance)
{
	VE_UNUSED(async);

	if (!vdHexBusy(instance->vdInstance) && instance->actionsCount && instance->actions[instance->actionsHead].id == id) {
		if (instance->actions[instance->actionsHead].actionType == VdVetiSet) {
#if CFG_WITH_LOG_PROTO
			switch (vetiGetType(instance->vetiInstance, id)) {
			case VETI_SN8:
			case VETI_UN8:
				logI(MODULE, "Received set hex item %04X of type VETI_SN8/UN8 with value %02X",
					 id, data[0]);
				break;
			case VETI_SN16:
			case VETI_UN16:
				logI(MODULE, "Received set hex item %04X of type VETI_SN16/UN16 with value %04X",
					 id, (((un16)data[1]) << 8) | (un16)data[0]);
				break;
			case VETI_SN32:
			case VETI_UN32:
				logI(MODULE, "Received set hex item %04X of type VETI_SN32/UN32 with value %08X",
					 id, (((un32)data[3]) << 24) | (((un32)data[2]) << 16) | (((un32)data[1]) << 8) | ((un32)data[0]));
				break;
			case VETI_PTR:
				logI(MODULE, "Received set hex item %04X of type VETI_PTR", id);
				break;
			case VETI_UNKNOWN:
				logE(MODULE, "Received set hex item %04X of type VETI_UNKNOWN", id);
				break;
			}
#endif
			/* We do not look at the flags here, because we are going to get the vreg again anyway.
			 * HEX_FLAG_UNKNOWN_ID -> this will also be flagged when we get.
			 * HEX_FLAG_PARAM_ERROR -> We do a get to get the actual value
			 * HEX_FLAG_READ_ONLY -> We do a get to get the actual value
			 * */
			setToGet(instance);
			return;
		} else {
			nextAction(instance);
		}
	}

	if (flags != 0) {
		const VeVariantValue v = vdGetInvalidValue(vetiToVeType(vetiGetType(instance->vetiInstance, id)));

		switch (flags) {
		case HEX_FLAG_UNKNOWN_ID:
			/* Invalidate the complete item as the connected device does not support it */
			vetiInvalidateById(instance->vetiInstance, instance->veDirectVetiHandle, id);
			return;
		case HEX_FLAG_READ_ONLY:
			/* We get this when getting a "Action" type item. */
			break;
		case HEX_FLAG_PARAM_ERROR:
			/* No idea when this could happen on a get. */
			logE(MODULE, "Unexpected param error on get of id %04X", id);
			break;
		default:
			logE(MODULE, "Unknown flags %02X for id %04X", flags, id);
			break;
		}
		/* Default action is to publish an invalid value */
		vetiPublish(instance->vetiInstance, instance->veDirectVetiHandle, id, &v);
		return;
	}

#if CFG_VEDIRECT_LAST_MESSAGE
	/* When extra info is turned on every get item request is stored in the last message item.
	 * This can be useful to reduce the amount of RAM used for storing items.
	 * The last message can be found here and is overwriten when a new get request is performed.
	 * The application should take care to copy the data if it should be memorized.
	 */
	if (length <= LAST_MESSAGE_LENGTH) {
		VeTinyItem * const item = vetiGetItem(instance->vetiInstance, VE_REG_LAST_MESSAGE);
		LastMessage * const lastMessage = item->value.Ptr;
		lastMessage->id = id;
		memcpy(lastMessage->data, data, length);
		item->properties.valid = veTrue;
		vetiPublishItem(instance->vetiInstance, instance->veDirectVetiHandle, VE_REG_LAST_MESSAGE, item);
	}
#endif
	switch (id) {
#if CFG_VEDIRECT_WITH_HISTORY || CFG_VEDIRECT_WITH_HISTORY_TOTAL
	case VE_REG_HISTORY_TOTAL:
	{
		if (length < 19 || (data[0] >= 1 && length < 20))
			return;
		VeTinyItem * const item = vetiGetItem(instance->vetiInstance, id);
		MpptTotalsHistory * const totals = item->value.Ptr;
		/* data[0] = extenstion, ignore for now */
		totals->extension = data[0];
		if (data[1] == 0) { // charger error
			totals->errors.chargerErrors[LAST_CHARGER_ERROR] = data[2];
			totals->errors.chargerErrors[PREV_CHARGER_ERROR] = data[3];
			totals->errors.chargerErrors[PREVPREV_CHARGER_ERROR] = data[4];
			totals->errors.chargerErrors[OLDEST_CHARGER_ERROR] = data[5];
		}
		totals->userYield = (((un32)data[9]) << 24) | (((un32)data[8]) << 16) | (((un32)data[7]) << 8) | (un32)data[6];
		totals->systemYield = (((un32)data[13]) << 24) | (((un32)data[12]) << 16) | (((un32)data[11]) << 8) | (un32)data[10];
		totals->maxPanelVoltage = (((un16)data[15]) << 8) | (un16)data[14];
		totals->maxBatteryVoltage = (((un16)data[17]) << 8) | (un16)data[16];
		totals->daysAvailable = data[18];
		if (totals->extension == 1)
			totals->minBatteryVoltage = (((un16)data[20]) << 8) | (un16)data[19];
		else
			totals->minBatteryVoltage = VE_INVALID_UN16;
		item->properties.valid = veTrue;
		vetiPublishItem(instance->vetiInstance, instance->veDirectVetiHandle, id, item);
		return;
	}
#endif
#if CFG_VEDIRECT_WITH_HISTORY || CFG_VEDIRECT_WITH_HISTORY_DAY0
	case VE_REG_HISTORY_DAY00:
#endif
#if CFG_VEDIRECT_WITH_HISTORY
	case VE_REG_HISTORY_DAY01:
	case VE_REG_HISTORY_DAY02:
	case VE_REG_HISTORY_DAY03:
	case VE_REG_HISTORY_DAY04:
	case VE_REG_HISTORY_DAY05:
	case VE_REG_HISTORY_DAY06:
	case VE_REG_HISTORY_DAY07:
	case VE_REG_HISTORY_DAY08:
	case VE_REG_HISTORY_DAY09:
	case VE_REG_HISTORY_DAY10:
	case VE_REG_HISTORY_DAY11:
	case VE_REG_HISTORY_DAY12:
	case VE_REG_HISTORY_DAY13:
	case VE_REG_HISTORY_DAY14:
	case VE_REG_HISTORY_DAY15:
	case VE_REG_HISTORY_DAY16:
	case VE_REG_HISTORY_DAY17:
	case VE_REG_HISTORY_DAY18:
	case VE_REG_HISTORY_DAY19:
	case VE_REG_HISTORY_DAY20:
	case VE_REG_HISTORY_DAY21:
	case VE_REG_HISTORY_DAY22:
	case VE_REG_HISTORY_DAY23:
	case VE_REG_HISTORY_DAY24:
	case VE_REG_HISTORY_DAY25:
	case VE_REG_HISTORY_DAY26:
	case VE_REG_HISTORY_DAY27:
	case VE_REG_HISTORY_DAY28:
	case VE_REG_HISTORY_DAY29:
	case VE_REG_HISTORY_DAY30:
#endif
#if CFG_VEDIRECT_WITH_HISTORY || CFG_VEDIRECT_WITH_HISTORY_DAY0
	{
		if (length != 34)
			return;
		VeTinyItem * const item = vetiGetItem(instance->vetiInstance, id);
		MpptDayHistory * const day = item->value.Ptr;
		/* data[0] = extenstion, ignore for now */
		day->yield = (((un32)data[4]) << 24) | (((un32)data[3]) << 16) | (((un32)data[2]) << 8) | (un32)data[1];
		day->consumed = (((un32)data[8]) << 24) | (((un32)data[7]) << 16) | (((un32)data[6]) << 8) | (un32)data[5];
		day->maxBatteryVoltage = (((un16)data[10]) << 8) | (un16)data[9];
		day->minBatteryVoltage = (((un16)data[12]) << 8) | (un16)data[11];
		if (data[13] == 0) { // charger error
			day->errors.chargerErrors[LAST_CHARGER_ERROR] = data[14];
			day->errors.chargerErrors[PREV_CHARGER_ERROR] = data[15];
			day->errors.chargerErrors[PREVPREV_CHARGER_ERROR] = data[16];
			day->errors.chargerErrors[OLDEST_CHARGER_ERROR] = data[17];
		}
		day->timeInBulk = (((un16)data[19]) << 8) | (un16)data[18];
		day->timeInAbsorption = (((un16)data[21]) << 8) | (un16)data[20];
		day->timeInFloat = (((un16)data[23]) << 8) | (un16)data[22];
		day->maxPower = (((un32)data[27]) << 24) | (((un32)data[26]) << 16) | (((un32)data[25]) << 8) | (un32)data[24];
		day->maxBatteryCurrent = (((un16)data[29]) << 8) | (un16)data[28];
		day->maxPanelVoltage = (((un16)data[31]) << 8) | (un16)data[30];
		day->daySequenceNumber = (((un16)data[33]) << 8) | (un16)data[32];
		item->properties.valid = veTrue;
		vetiPublishItem(instance->vetiInstance, instance->veDirectVetiHandle, id, item);
		return;
	}
#endif
	case VE_REG_SERIAL_NUMBER: {
		if (length != 11)
			return;

		VeTinyItem * const item = vetiGetItem(instance->vetiInstance, id);
		memcpy(item->value.Ptr, data, 11);
		item->properties.valid = veTrue;
		vetiPublishItem(instance->vetiInstance, instance->veDirectVetiHandle, id, item);
		return;
	}

	default:
		break;
	}

	VeVariantValue value;
	value.LARGE_FIELD = 0;
	switch (vetiGetType(instance->vetiInstance, id)) {
	case VETI_SN8:
	case VETI_UN8:
		value.UN8 = data[0];
		logI(MODULE, "Received hex item %04X of type VETI_SN8/UN8 with value %02X", id, value.UN8);
		break;
	case VETI_SN16:
	case VETI_UN16:
		value.UN16 = (((un16)data[1]) << 8) | (un16)data[0];
		logI(MODULE, "Received hex item %04X of type VETI_SN16/UN16 with value %04X", id, value.UN16);
		break;
	case VETI_SN32:
	case VETI_UN32:
		value.UN32 = (((un32)data[3]) << 24) | (((un32)data[2]) << 16) | (((un32)data[1]) << 8) | ((un32)data[0]);
		logI(MODULE, "Received hex item %04X of type VETI_SN32/UN32 with value %08X", id, value.UN32);
		break;
	case VETI_PTR:
		value.CPtr = data;
		logI(MODULE, "Received hex item %04X of type VETI_PTR", id);
		break;
	case VETI_UNKNOWN:
		logE(MODULE, "Received hex item %4.4X of type VETI_UNKNOWN", id);
		return;
	}
	vetiPublish(instance->vetiInstance, instance->veDirectVetiHandle, id, &value);
}

void vdVetiHexErrorNotify(const VdHexError error, VdVetiInstance * instance)
{
	if (!instance->actionsCount)
		return;

	const VdVetiAction * const action = &instance->actions[instance->actionsHead];

	switch (error) {
	case VdHexTimeout:
		logW(MODULE, "Received VdHexTimeout. Retrying.");
		performAction(instance);
		break;

	case VdHexChecksum:
		/* Communication failure in response. We will get a timeout after a while. */
	case VdHexRxBufferOverflow:
		/* It might be that the vedirect stack missed the \n and the data just kept coming in.
		 * Lets just wait for the timeout and retry with the chance that the returned message
		 * is always too large. */
		break;

	case VdHexBusy:
		/* This is clearly an error on our side (or the application because there should not be 2 requests
		 * active at the same time. */
		logE(MODULE, "Received a VdHexBusy. actionHead: %d, actionCount: %d, actionType: %s, id: %04X",
			 instance->actionsHead, instance->actionsCount, actionTypeStr(action->actionType), action->id);
		break;
	case VdHexSetTypeUnsupported:
		logE(MODULE, "Received a VdHexSetTypeUnsupported. actionHead: %d, actionCount: %d, actionType: %s, id: %04X",
			 instance->actionsHead, instance->actionsCount, actionTypeStr(action->actionType), action->id);
		break;
	case VdHexTxBufferOverflow:
		logE(MODULE, "Received a VdHexTxBufferOverflow. actionHead: %d, actionCount: %d, actionType: %s, id: %04X",
			 instance->actionsHead, instance->actionsCount, actionTypeStr(action->actionType), action->id);
		break;
	}
}

static void vetiItemChanged(void * context, const VetiId id, const VeVariantValue * value, veBool valid)
{
	VdVetiInstance * const instance = context;

	logI(MODULE, "Item changed %04X %X %d", id, value->UN32, valid);

	/* This was an invalidate of the item */
	if (!valid)
		return;

	if (signalActionsFull(instance, VdVetiSet, id)) {
		return;
	}

	logI(MODULE, "Adding VdVetiSet for %04X to actionList: %d %d", id, instance->actionsCount, instance->actionsHead);

	VdVetiAction * const action = &instance->actions[actionIdx(instance->actionsHead, instance->actionsCount)];
	action->actionType = VdVetiSet;
	action->id = id;
	action->value = *value;

	if (!instance->actionsCount++)
		performAction(instance);
}

static void vetiPublisher(VdVetiInstance * instance, const VetiId id)
{
	if (signalActionsFull(instance, VdVetiGet, id)) {
		return;
	}

	logI(MODULE, "Adding VdVetiGet for %04X to actionList: %d %d", id, instance->actionsCount, instance->actionsHead);
	VdVetiAction * const action = &instance->actions[actionIdx(instance->actionsHead, instance->actionsCount)];
	action->actionType = VdVetiGet;
	action->id = id;

	if (!instance->actionsCount++)
		performAction(instance);
}

void vdVetiInit(VdVetiInstance * instance, VeDirectInstance * vd, VetiInstance * veti, void * context)
{
	memset(instance, 0, sizeof(*instance));
	instance->context = context;
	instance->vdInstance = vd;
	instance->vetiInstance = veti;
	instance->actionsHead = 0;
	instance->actionsCount = 0;

	instance->veDirectVetiHandle = vetiConnect(veti);
	vdSetContext(vd, instance);
	vdConnect(vd, TextItemNotify, vdVetiTextItemNotify);
	vdConnect(vd, HexMessageNotify, vdVetiHexMessageNotify);
	vdConnect(vd, StatusChange, vdVetiStatusChange);
	vdConnect(vd, HexErrorNotify, vdVetiHexErrorNotify);
	vetiConnectPublischer(veti, (PubReqFunc)vetiPublisher, instance);
	vetiConnectSubscriber(veti, instance->veDirectVetiHandle, vetiItemChanged, instance);
	vetiSubscribeAll(veti, instance->veDirectVetiHandle);
}


