#include <velib/base/base.h>
#include <velib/utils/ve_logger.h>
#include <string.h>

#include "vedirect.h"
#include "velib/ve_regs.h"
#include "vehex.h"

#define MODULE "VE.Hex"

#define HEX_CHECKSUM_OK			0x55
#define HEX_CHECKSUM_SEED		0x55
#define HEX_EMPTY_VALUE			0xAAAA

#define NibbleToBin(a)	( ((a)>='0' && (a)<='9') ?  (a)-'0' : ( ((a)>='A' && (a)<='F') ? ((a)-'A')+10 : 0xFF ) )
#define binToHighNibble(b)	hex[((b)&0xF0)>>4]
#define binToLowNibble(b)	hex[(b)&0x0F]

typedef enum { FlagUnknowdId = 0x01, FlagNotSupported = 0x02, FlagParamError = 0x04 } Flags;

static un8 const hex[] = { '0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F' };

/*
 * Local functions prototypes
 */
static un8 veHexBuildMessage(VdHexInstance * instance, CommandValues command, un16 id, un8 flags, const un8 * payload, un8 length);
static void veHexSendMessage(VdHexInstance * instance, un8 const * data, un8 length);

/*
 * Local functions
 */

static un8 veHexBuildMessage(VdHexInstance * instance, const CommandValues command, const un16 id, const un8 flags, const un8 * const payload, const un8 length)
{
	/* Message format
	 * :[command][payload][checksum]\n
	 */
	un8 i;
	un8 j = 2;
	un8 checksum = HEX_CHECKSUM_SEED;

	instance->txBuf[0] = ':';
	instance->txBuf[1] = hex[command];
	checksum -= command;
	switch (command) {
	case CmdVal_EnterBoot_Reserved:
		//strcpy(txBuf+2, "51FA51FA51FA51FA51FA");
		break;
	case CmdVal_Get_GetResponse:
	case CmdVal_Set_SetResponse:
		/* id */
		instance->txBuf[2] = binToHighNibble(id);
		instance->txBuf[3] = binToLowNibble(id);
		checksum -= id & 0xFF;
		instance->txBuf[4] = binToHighNibble(id>>8);
		instance->txBuf[5] = binToLowNibble(id>>8);
		checksum -= (id>>8) & 0xFF;
		/* Flags */
		instance->txBuf[6] = binToHighNibble(flags);
		instance->txBuf[7] = binToLowNibble(flags);
		checksum -= flags;
		j += 6;
		/* Fall though */
	case CmdVal_Ping_Done:
	case CmdVal_AppVersion_UnknownCommand:
	case CmdVal_DeviceId_MiscError:
	case CmdVal_Restart_VerifyError:
	case CmdVal_Program_PingResponse:
		if (j + 3 + length * 2 > sizeof(instance->txBuf)) {
			logE(MODULE, "Overflow in tx buffer: %X %d", command, length);
			vdHexErrorNotify(instance, VdHexTxBufferOverflow);
			return 0;
		}
		for (i = 0; i < length; i++) {
			instance->txBuf[j++] = binToHighNibble(payload[i]);
			instance->txBuf[j++] = binToLowNibble(payload[i]);
			checksum -= payload[i];
		}
		instance->txBuf[j++] = binToHighNibble(checksum);
		instance->txBuf[j++] = binToLowNibble(checksum);
		instance->txBuf[j++] = '\n';
		return j;
	case CmdVal_Erase_ChecksumError:
	case CmdVal_NA_Notify:
		break;
	}
	return 0;
}

static un8 veHexBuildCommand(VdHexInstance * instance, CommandValues command)
{
	return veHexBuildMessage(instance, command, 0, 0, NULL, 0);
}

/*
 * vedirect API
 */
void veHexInit(VdHexInstance * instance)
{
	memset(instance, 0, sizeof(*instance));
	instance->message.data = instance->rxBuf;
}

static void veHexSendMessage(VdHexInstance * instance, un8 const * const data, const un8 length)
{
	vdTxData(instance, data, length);
}

/*
 * API functions
 */
void veHexSendPing(VdHexInstance * instance)
{
	const un8 length = veHexBuildCommand(instance, CmdVal_Ping_Done);
	veHexSendMessage(instance, instance->txBuf, length);
}

void veHexGetDeviceId(VdHexInstance * instance)
{
	un8 length = veHexBuildCommand(instance, CmdVal_DeviceId_MiscError);
	veHexSendMessage(instance, instance->txBuf, length);
}

void veHexRestart(VdHexInstance * instance)
{
	un8 length = veHexBuildCommand(instance, CmdVal_Restart_VerifyError);
	veHexSendMessage(instance, instance->txBuf, length);
}

void veHexGet(VdHexInstance * instance, const un16 id)
{
	un8 mesgLength = veHexBuildMessage(instance, CmdVal_Get_GetResponse, id, 0, NULL, 0);
	veHexSendMessage(instance, instance->txBuf, mesgLength);
}

void veHexSet(VdHexInstance * instance, un16 const id, un8 const * const payload, un8 const length)
{
	un8 mesgLength = veHexBuildMessage(instance, CmdVal_Set_SetResponse, id, 0, payload, length);
	veHexSendMessage(instance, instance->txBuf, mesgLength);
}

/*
 *  Frame Handler API
 */
veBool veHexRxEvent(VdHexInstance * instance, un8 c)
{
	if (instance->leastNibble) {
		instance->byte |= NibbleToBin(c);
		if (instance->byteCount < sizeof(instance->rxBuf)) {
			instance->rxBuf[instance->byteCount++] = instance->byte;
			instance->checksum += instance->byte;
		} else {
			logE(MODULE,"Rx buffer overflow: %i", instance->byteCount);
			vdHexErrorNotify(instance, VdHexRxBufferOverflow);
			instance->rxState = StateIgnore;
		}
	} else {
		instance->byte = NibbleToBin(c) << 4;
	}
	instance->leastNibble = !instance->leastNibble;

	switch (instance->rxState) {
	case StateCommand:
		switch (c) {
		case CmdChar_AppVersion_UnknownCommand:
		case CmdChar_Erase_ChecksumError:
		case CmdChar_Restart_VerifyError:
			instance->rxState = StateIgnore;
			break;

		case CmdChar_Program_PingResponse:
			instance->message.veRegId = VE_REG_APP_VER;
			/* Fall through */
		case CmdChar_EnterBoot_Reserved:
		case CmdChar_DeviceId_MiscError:
		case CmdChar_Ping_Done:
			instance->rxState = StateData;
			break;

		case CmdChar_NA_Notify:
		case CmdChar_Get_GetResponse:
		case CmdChar_Set_SetResponse:
			instance->rxState = StateVeRegId;
			break;

		default:
			logE(MODULE, "Unknown command %X received", c);
			instance->rxState = StateIgnore;
			break;
		}
		instance->checksum = NibbleToBin(c);
		instance->byteCount = 0;
		instance->leastNibble = veFalse;
		instance->message.response = c;
		instance->message.flags = 0;
		break;
	case StateIgnore:
		instance->byteCount = 0;
		if (c == '\n') {
			instance->rxState = StateCommand;
			return veTrue;
		}
		break;
	case StateVeRegId:
		if (instance->byteCount == 2) {
			instance->message.veRegId =  (instance->rxBuf[1] << 8) | instance->rxBuf[0];
			instance->byteCount = 0;
			instance->rxState = StateFlags;
		}
		break;
	case StateFlags:
		if (instance->byteCount == 1) {
			instance->message.flags =  instance->rxBuf[0];
			instance->byteCount = 0;
			instance->rxState = StateData;
		}
		break;
	case StateData:
		switch (c) {
		case '\n':
#if CFG_WITH_LOG_PROTO
			{
				char buffer[sizeof(instance->rxBuf) * 2 + 20];
				int idx = 0;
				int i;
				buffer[idx++] = ':';
				buffer[idx++] = instance->message.response;
				switch (instance->message.response) {
				case CmdChar_NA_Notify:
				case CmdChar_Get_GetResponse:
				case CmdChar_Set_SetResponse:
					buffer[idx++] = hex[(instance->message.veRegId >> 4) & 0xF];
					buffer[idx++] = hex[(instance->message.veRegId >> 0) & 0xF];
					buffer[idx++] = hex[(instance->message.veRegId >> 12) & 0xF];
					buffer[idx++] = hex[(instance->message.veRegId >> 8) & 0xF];

					buffer[idx++] = hex[(instance->message.flags >> 4) & 0xF];
					buffer[idx++] = hex[(instance->message.flags >> 0) & 0xF];
					break;
				default:
					break;
				}
				for (i = 0; i < instance->byteCount; ++i) {
					buffer[idx++] = hex[(instance->rxBuf[i] >> 4) & 0xF];
					buffer[idx++] = hex[(instance->rxBuf[i] >> 0) & 0xF];
				}
				buffer[idx++] = 0;
				logI(MODULE, "Received: %s", buffer);
			}
#endif
			/* First check if the checksum is ok so we can assume we got a correct message */
			if ((instance->checksum & 0xFF) == HEX_CHECKSUM_OK) {
				/* Remove the checksum from the byte count*/
				instance->byteCount--;
				switch (instance->message.response) {
				case CmdChar_EnterBoot_Reserved:
				{
					static const un8 enterBootPayload[] =
						{0x51, 0xFA, 0x51, 0xFA, 0x51, 0xFA, 0x51, 0xFA, 0x51, 0xFA};

					if (instance->byteCount != 10) {
						logE(MODULE, "Got EnterBoot/Reserved command with data length %d", instance->byteCount);
					} else if (memcmp(instance->rxBuf, enterBootPayload, 10) != 0) {
						logE(MODULE, "Invalid payload in Enter_Boot message");
					} else {
						logI(MODULE, "Valid Enter_Boot message");
						vdRequestNotify(instance, CmdChar_EnterBoot_Reserved);
					}
					break;
				}
				case CmdChar_Ping_Done:
					if (instance->byteCount != 0) {
						logE(MODULE, "Got a Ping/Done message with data length %d", instance->byteCount);
					} else {
						logI(MODULE, "Received valid ping request");
						const un16 response = vdRequestNotify(instance, CmdChar_Ping_Done);
						const un8 mesgLength = veHexBuildMessage(instance, CmdVal_Program_PingResponse, 0, 0, (un8 const *)&response, 2);
						veHexSendMessage(instance, instance->txBuf, mesgLength);
					}
					break;
				case CmdChar_DeviceId_MiscError:
					if (instance->byteCount != 0) {
						logE(MODULE, "Got a DeviceId/MiscError message with data length %d", instance->byteCount);
					} else {
						logI(MODULE, "Received DeviceID request");
						const un16 response = vdRequestNotify(instance, CmdChar_DeviceId_MiscError);
						const un8 mesgLength = veHexBuildMessage(instance, CmdVal_Program_PingResponse, 0, 0, (un8 const *)&response, 2);
						veHexSendMessage(instance, instance->txBuf, mesgLength);
					}
					break;
				case CmdChar_Program_PingResponse:
				case CmdChar_NA_Notify:
				case CmdChar_Get_GetResponse:
				case CmdChar_Set_SetResponse:
					if (instance->message.flags & HEX_FLAG_UNKNOWN_ID) {
						logW(MODULE,"unknown id failure: %x", instance->message.veRegId);
					} else if (instance->message.flags & HEX_FLAG_PARAM_ERROR) {
						logE(MODULE,"param failure: %x", instance->message.veRegId);
					} else if (instance->message.flags & HEX_FLAG_READ_ONLY) {
						logE(MODULE,"readonly failure: %x", instance->message.veRegId);
					}
					instance->message.length = instance->byteCount;
					vdHexItemUpdate(instance, &instance->message);
					break;
				default:
					logE(MODULE,"Stating error: In StateData with message.response being %X", instance->message.response);
					break;
				}
			} else {
				logE(MODULE,"checksum failure: %x", instance->checksum & 0xFF);
				vdHexErrorNotify(instance, VdHexChecksum);
			}
			instance->byteCount = 0;
			instance->rxState = StateCommand;
			return veTrue;
		default:
			break;
		}
		break;
	}

	return veFalse;
}
