#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <velib/utils/ve_logger.h>
#include <velib/vecan/products.h>
#include "vetext.h"
#include "vedirect.h"

#define MODULE			"VE.Text"
#define ALARM_ON_POS	0x8000


#define CACHE_INDEX(item) (offsetof(VdTagValues, named.item) / sizeof(VeTagCache))

/* This list should be ASCII sorted on the tag */
static const VeTagRecord veTextItems[] =
{
#if CFG_VETEXT_WITH_INVERTER
	{ "AC_OUT_I", VE_REG_AC_OUT_CURRENT,                CACHE_INDEX(acouti), 1,   1,    0, VE_SN16 },
	{ "AC_OUT_V", VE_REG_AC_OUT_VOLTAGE,                CACHE_INDEX(acoutv), 1,   1,    0, VE_UN16 },
#endif
#if CFG_VETEXT_WITH_BMV
	{ "ALARM", VE_REG_ALARM_STATUS,                     CACHE_INDEX(alarm),  0,    0,    0, VE_UN8 },
#endif
#if CFG_VETEXT_WITH_BMV || CFG_VETEXT_WITH_INVERTER
	{ "AR",    VE_REG_ALARM_REASON,                     CACHE_INDEX(ar),     0,    0,    0, VE_UN16 },
#endif
	#if CFG_VETEXT_WITH_BMV
	{ "BMV",   VE_REG_BMV,                              CACHE_INDEX(bmv), 	0,    0, BMV_LENGTH, VE_STR  },
	{ "CE",    VE_REG_CAH,                              CACHE_INDEX(cah),  100,    1,   50, VE_SN32 },
#endif
#if CFG_VETEXT_WITH_MPPT || CFG_VETEXT_WITH_INVERTER
	{ "CS",    VE_REG_DEVICE_STATE,                     CACHE_INDEX(cs),     1,    1,    0, VE_UN8  },
#endif
#if CFG_VETEXT_WITH_BMV
	{ "DM",    VE_REG_BATTERY_MID_POINT_DEVIATION,      CACHE_INDEX(dm),     1,    1,    0, VE_SN16 },
#endif
#if CFG_VETEXT_WITH_MPPT
	{ "ERR",   VE_REG_CHR_ERROR_CODE,                   CACHE_INDEX(err),    1,    1,    0, VE_UN8  },
#endif
	{ "FW",    VE_REG_APP_VER,                          CACHE_INDEX(fw),     1,    1,    0, VE_UN16 },
#if CFG_VETEXT_WITH_BMV
	{ "H1",    VE_REG_HIST_DEEPEST_DISCHARGE,           CACHE_INDEX(h1),   100,    1,    0, VE_SN32 },
	{ "H10",   VE_REG_HIST_NR_OF_AUTO_SYNCS,            CACHE_INDEX(h10),    1,    1,    0, VE_UN32 },
	{ "H11",   VE_REG_HIST_NR_OF_LOW_VOLTAGE_ALARMS,    CACHE_INDEX(h11),    1,    1,    0, VE_UN32 },
	{ "H12",   VE_REG_HIST_NR_OF_HIGH_VOLTAGE_ALARMS,   CACHE_INDEX(h12),    1,    1,    0, VE_UN32 },
	{ "H13",   VE_REG_HIST_NR_OF_LOW_VOLTAGE_2_ALARMS,  CACHE_INDEX(h13),    1,    1,    0, VE_UN32 },
	{ "H14",   VE_REG_HIST_NR_OF_HIGH_VOLTAGE_2_ALARMS, CACHE_INDEX(h14),    1,    1,    0, VE_UN32 },
	{ "H15",   VE_REG_HIST_MIN_VOLTAGE_2,               CACHE_INDEX(h15),   10,    1,    0, VE_SN32 },
	{ "H16",   VE_REG_HIST_MAX_VOLTAGE_2,               CACHE_INDEX(h16),   10,    1,    0, VE_SN32 },
	{ "H17",   VE_REG_HIST_KWH_OUT,                     CACHE_INDEX(h17),    1,    1,    0, VE_UN32 },
	{ "H18",   VE_REG_HIST_KWH_IN,                      CACHE_INDEX(h18),    1,    1,    0, VE_UN32 },
#endif
#if CFG_VETEXT_WITH_MPPT
	{ "H19",   VE_REG_CHR_USER_YIELD,                   CACHE_INDEX(h19),    1,    1,    0, VE_UN32 },
#endif
#if CFG_VETEXT_WITH_BMV
	{ "H2",    VE_REG_HIST_LAST_DISCHARGE,              CACHE_INDEX(h2),   100,    1,    0, VE_SN32 },
#endif
#if CFG_VETEXT_WITH_MPPT
	{ "H20",   VE_REG_CHR_TODAY_YIELD,                  CACHE_INDEX(h20),    1,    1,    0, VE_UN32 },
	{ "H21",   VE_REG_CHR_TODAY_PMAX,                   CACHE_INDEX(h21),    1,    1,    0, VE_UN32 },
	{ "H22",   VE_REG_CHR_YESTERDAY_YIELD,              CACHE_INDEX(h22),    1,    1,    0, VE_UN32 },
	{ "H23",   VE_REG_CHR_YESTERDAY_PMAX,               CACHE_INDEX(h23),    1,    1,    0, VE_UN32 },
#endif
#if CFG_VETEXT_WITH_BMV
	{ "H3",    VE_REG_HIST_AVERAGE_DISCHARGE,           CACHE_INDEX(h3),   100,    1,    0, VE_SN32 },
	{ "H4",    VE_REG_HIST_NR_OF_CHARGE_CYCLES,         CACHE_INDEX(h4),     1,    1,    0, VE_UN32 },
	{ "H5",    VE_REG_HIST_NR_OF_FULL_DISCHARGES,       CACHE_INDEX(h5),     1,    1,    0, VE_UN32 },
	{ "H6",    VE_REG_HIST_CUMULATIVE_AH,               CACHE_INDEX(h6),   100,    1,    0, VE_SN32 },
	{ "H7",    VE_REG_HIST_MIN_VOLTAGE,                 CACHE_INDEX(h7),    10,    1,    0, VE_UN32 },
	{ "H8",    VE_REG_HIST_MAX_VOLTAGE,                 CACHE_INDEX(h8),    10,    1,    0, VE_UN32 },
	{ "H9",    VE_REG_HIST_SECS_SINCE_LAST_FULL_CHARGE, CACHE_INDEX(h9),     1,    1,    0, VE_UN32 },
#endif
#if CFG_VETEXT_WITH_MPPT
	{ "HSDS",  VE_REG_HSDS,                             CACHE_INDEX(hsds),   1,    1,    0, VE_UN16 },
#endif
#if CFG_VETEXT_WITH_BMV || CFG_VETEXT_WITH_MPPT
	{ "I",     VE_REG_DC_CHANNEL1_CURRENT,              CACHE_INDEX(i),    100,    1,   50, VE_SN16 },
#endif
#if CFG_VETEXT_WITH_MPPT
	{ "IL",    VE_REG_DC_OUTPUT_CURRENT,                CACHE_INDEX(il),   100,    1,   50, VE_UN16 },
	{ "LOAD",  VE_REG_DC_OUTPUT_STATUS,                 CACHE_INDEX(load),   1,    1,    0, VE_UN8  },
#endif
#if CFG_VETEXT_WITH_INVERTER
	{ "MODE",  VE_REG_DEVICE_MODE,                      CACHE_INDEX(mode),   1,    1,    0, VE_UN8  },
#endif
#if CFG_VETEXT_WITH_BMV
	{ "P",     VE_REG_DC_CHANNEL1_POWER,                CACHE_INDEX(p),      1,    1,    0, VE_SN16 },
#endif
	{ "PID",   VE_REG_PRODUCT_ID,                       CACHE_INDEX(pid),    1,    1,    0, VE_UN16 },
#if CFG_VETEXT_WITH_MPPT
	{ "PPV",   VE_REG_DC_INPUT_POWER,                   CACHE_INDEX(ppv),    1,  100,    0, VE_UN32 },
#endif
#if CFG_VETEXT_WITH_BMV || CFG_VETEXT_WITH_MPPT
	{ "PROD",  VE_REG_PROD,                             CACHE_INDEX(prod),   1,    1, PROD_LENGTH, VE_STR  },
	{ "RELAY", VE_REG_RELAY_CONTROL,                    CACHE_INDEX(relay),  0,    0,    0, VE_UN8  },
#endif
	{ "SER#",  VE_REG_SERIAL_NUMBER,                    CACHE_INDEX(ser),    0,    0, SER_LENGTH, VE_STR  },
#if CFG_VETEXT_WITH_BMV
	{ "SOC",   VE_REG_SOC,                              CACHE_INDEX(soc),    1,   10,    0, VE_UN16 },
	{ "T",     VE_REG_BAT_TEMPERATURE,                  CACHE_INDEX(t),      0,    0,    0, VE_UN16 },
	{ "TTG",   VE_REG_TTG,                              CACHE_INDEX(ttg),    0,    0,    0, VE_UN16 },
#endif
	{ "V",     VE_REG_DC_CHANNEL1_VOLTAGE,              CACHE_INDEX(v),     10,    1,    5, VE_UN16 },
#if CFG_VETEXT_WITH_BMV || CFG_VETEXT_WITH_MPPT
	{ "VER#",  VE_REG_VER,                              CACHE_INDEX(ver),    0,    0,    0, VE_UN16 },
#endif
#if CFG_VETEXT_WITH_BMV
	{ "VM",    VE_REG_BATTERY_MID_POINT_VOLTAGE,        CACHE_INDEX(vm),    10,    1,    5, VE_UN16 },
#endif
#if CFG_VETEXT_WITH_MPPT
	{ "VPV",   VE_REG_DC_INPUT_VOLTAGE,                 CACHE_INDEX(vpv),   10,    1,    5, VE_UN16 },
#endif
#if CFG_VETEXT_WITH_BMV
	{ "VS",    VE_REG_DC_CHANNEL2_VOLTAGE,              CACHE_INDEX(vs),    10,    1,    5, VE_UN16 },
#endif
#if CFG_VETEXT_WITH_INVERTER
	{ "WARN", VE_REG_WARNING_REASON,                    CACHE_INDEX(warn),   1,    1,    0, VE_UN8  },
#endif
};
static const un8 numOfItems = sizeof(veTextItems)/sizeof(VeTagRecord);

#define ASSERT_CONCAT_(a, b) a##b
#define ASSERT_CONCAT(a, b) ASSERT_CONCAT_(a, b)
#define STATIC_ASSERT(e,m) \
	typedef char ASSERT_CONCAT(assert_array_line_, __LINE__)[(e) ? 1 : -1]

STATIC_ASSERT(sizeof(veTextItems)/sizeof(VeTagRecord) == sizeof(VdTagValues)/sizeof(VeTagCache),"veTextItems should have the same length as VdTagValues");

/*
 * Local functions
 */
static int searchCompare(const void * a, const void * b)
{
	const char * key = a;
	const VeTagRecord *item = b;
	return strcmp(key,item->tagName);
}

/*
 * API functions
 */
void veTextInit(VdTextInstance * instance)
{
	memset(instance, 0, sizeof(*instance));
#if CFG_VETEXT_WITH_BMV
	instance->tagValues.named.bmv.value.Ptr = instance->bmv;
#endif
#if CFG_VETEXT_WITH_BMV || CFG_VETEXT_WITH_MPPT
	instance->tagValues.named.prod.value.Ptr = instance->prod;
	instance->tagValues.named.prod.hex = veTrue;
	instance->tagValues.named.ver.hex = veTrue;
#endif
	instance->tagValues.named.ser.value.Ptr = instance->ser;
	instance->tagValues.named.fw.hex = veTrue;
	instance->tagValues.named.pid.hex = veTrue;
	instance->tagValues.named.pid.value.UN16 = VE_PROD_NOT_SET;

}

#if UINTPTR_MAX > UINT32_MAX
#define LARGE_FIELD Ptr
#else
#define LARGE_FIELD UN32
#endif

void veTextRxEvent(VdTextInstance * instance, char const * const name, char const * const value)
{
	VeTagRecord *item = bsearch(name, veTextItems, numOfItems, sizeof(VeTagRecord), searchCompare);
	if (item == NULL) {
		logW(MODULE, "Tag not found: %s=%s", name, value);
		return;
	}
	VeVariantValue tmpValue;
	veBool valid = veTrue;
	sn32 val;

	logI(MODULE,"%s=%s", name, value);

	tmpValue.LARGE_FIELD = 0;
	if (strcmp(value,"---") == 0) {
		valid = veFalse;
	} else {
		switch (item->regId) {
#if CFG_VETEXT_WITH_BMV || CFG_VETEXT_WITH_MPPT
		case VE_REG_VER:
			{
			char * end;
			un8  major =  (un8)strtol(value, &end, 10);
			un8  minor =  (un8)strtol(end+1, NULL, 10);
			tmpValue.UN16 = major<<8 | minor;
			if (tmpValue.UN16 != instance->tagValues.named.fw.value.UN16) {
				instance->tagValues.named.fw.value.UN16 = tmpValue.UN16;
				instance->tagValues.named.fw.updated = veTrue;
				instance->tagValues.named.fw.valid = veTrue;
			}
			return;
			}
#endif
#if CFG_VETEXT_WITH_BMV || CFG_VETEXT_WITH_MPPT
		case VE_REG_RELAY_CONTROL:
#if CFG_VETEXT_WITH_MPPT
		case VE_REG_DC_OUTPUT_STATUS:
#endif
#if CFG_VETEXT_WITH_BMV
		case VE_REG_ALARM_STATUS:
#endif
			if (strcmp(value,"ON") == 0)
				tmpValue.UN8 = 1;
			else if (strcmp(value,"OFF") == 0)
				tmpValue.UN8 = 0;
			else {
				valid = veFalse;
				break;
			}
#if CFG_VETEXT_WITH_BMV
			if (item->regId == VE_REG_ALARM_STATUS) {
				un16 oldAlarm = instance->tagValues.named.ar.value.UN16;
				if (tmpValue.UN8 == 1)
					instance->tagValues.named.ar.value.UN16 |= ALARM_ON_POS;
				else
					instance->tagValues.named.ar.value.UN16 &= (~ALARM_ON_POS);
				if (!instance->tagValues.named.ar.valid
						|| instance->tagValues.named.ar.value.UN16 != oldAlarm)
					instance->tagValues.named.ar.updated = veTrue;
				instance->tagValues.named.ar.valid = veTrue;

				return; /* VE_REG_ALARM_STATUS itself is not published */
			}
#endif
			break;
#endif
#if CFG_VETEXT_WITH_BMV
		case VE_REG_TTG:
			val = (sn32)strtol(value, NULL, 10);
			if (val > 0)
				tmpValue.UN16 = (un16)val;
			else
				valid = veFalse;
			break;
		case VE_REG_BAT_TEMPERATURE:
			val = (sn32)strtol(value, NULL, 10);
			tmpValue.UN16 = (un16)(val*100+27315); /* C -> K */
			break;
#endif
#if CFG_VETEXT_WITH_INVERTER
		case VE_REG_WARNING_REASON:
#endif
#if CFG_VETEXT_WITH_BMV || CFG_VETEXT_WITH_INVERTER
		case VE_REG_ALARM_REASON:
			val = (sn32)strtol(value, NULL, 10);
			if (val) val |= ALARM_ON_POS;
			tmpValue.UN16 = (un16)val;
			break;
#endif
		default:
			if (instance->tagValues.array[item->cacheIndex].hex)
				val = (sn32)strtoul(value, NULL, 16);
			else
				val = (sn32)strtol(value, NULL, 10);
			if (val < 0)
				val -= item->rounder;
			else
				val += item->rounder;
			switch (item->type) {
			case VE_SN8:
				tmpValue.SN8 = (sn8) (val * item->multiplier / item->divider);
				break;
			case VE_UN8:
				tmpValue.UN8 = (un8) (val * item->multiplier / item->divider);
				break;
			case VE_SN16:
				tmpValue.SN16 = (sn16) (val * item->multiplier / item->divider);
				break;
			case VE_UN16:
				tmpValue.UN16 = (un16) (val * item->multiplier / item->divider);
				break;
			case VE_SN32:
				tmpValue.SN32 = (sn32) (val * item->multiplier / item->divider);
				break;
			case VE_UN32:
				tmpValue.UN32 = (un32) (val * item->multiplier / item->divider);
				break;
			case VE_STR:
				if (!instance->tagValues.array[item->cacheIndex].valid ||
						strcmp(instance->tagValues.array[item->cacheIndex].value.Ptr, value)) {
					strncpy(instance->tagValues.array[item->cacheIndex].value.Ptr, value, item->rounder - 1);
					instance->tagValues.array[item->cacheIndex].valid = veTrue;
					instance->tagValues.array[item->cacheIndex].updated = veTrue;
				}
				return;
			default:
				break;
			}
		}
	}
	if (!valid)
		tmpValue = vdGetInvalidValue(item->type);

	if (!instance->tagValues.array[item->cacheIndex].valid  /* Always forward tag that was not yet found */
			|| instance->tagValues.array[item->cacheIndex].value.LARGE_FIELD != tmpValue.LARGE_FIELD) /* Or when the value has changed */
	{
		instance->tagValues.array[item->cacheIndex].value.LARGE_FIELD = tmpValue.LARGE_FIELD;
		instance->tagValues.array[item->cacheIndex].valid = veTrue;
		instance->tagValues.array[item->cacheIndex].updated = veTrue;
	}
}

/* This function is called when the end of a frame is detected */
void veTextFrameEndEvent(VdTextInstance * instance, veBool valid)
{
	vdFrameReceived(instance, valid);
	if (valid) {
		un8 flags;
		do {
			un8 i;
			flags = 0;
			for (i = 0; i < numOfItems; i++) {
				const VeTagRecord *item = veTextItems+i;
				VeTagCache *cache = &instance->tagValues.array[item->cacheIndex];
				if (cache->updated) {
					if (vdTextItemUpdate(instance, item->regId, &cache->value, item->type)) {
						cache->updated = veFalse;
						flags |= 2;
					} else {
						flags |= 1;
					}
				}
			}
		} while (flags == 3);
	}
}
