#include <velib/utils/ve_logger.h>
#include <velib/platform/console.h>

#include "vedirect_service.h"
#include "items_helpers.h"

/* This file should support multiple instances of the vedirect_service and should
 * thus NOT contain any non-const global variables */

#define MODULE	"bmv"

#define ALARM_LOW_VOLTAGE		0x01		/* low voltage alarm */
#define ALARM_HIGH_VOLTAGE		0x02		/* high voltage alarm */
#define ALARM_LOW_SOC			0x04		/* low SOC alarm */
#define ALARM_LOW_VOLTAGE2		0x08		/* low voltage2 alarm */
#define ALARM_HIGH_VOLTAGE2		0x10		/* high voltage2 alarm */
#define ALARM_LOW_TEMPERATURE	0x20		/* low temperature alarm */
#define ALARM_HIGH_TEMPERATURE	0x40		/* high temperature alarm */
#define ALARM_MID_VOLTAGE		0x80		/* mid voltage alarm */

typedef enum
{
	DBUS_ALARM_NONE,
	DBUS_ALARM_WARNNG,
	DBUS_ALARM_ERROR
} DBusAlarm;

typedef struct
{
	size_t item;
	un16 mask;
} AlarmDef;

static ItemInfo const itemInfo[] =
{
#ifdef ITEM_TREE
	{VE_REG_ALARM_STATUS,						item_offset(bmv.alarms.alarm),						IOPT( 0,		"Alarms/Alarm",						&none,		0),	}, /* Generic alarm (deprecated) */
	{VE_REG_INVALID,							item_offset(bmv.alarms.lowVoltage),					IOPT( 0,		"Alarms/LowVoltage",				&none,		0),	},
	{VE_REG_INVALID,							item_offset(bmv.alarms.highVoltage),				IOPT( 0,		"Alarms/HighVoltage",				&none,		0),	},
	{VE_REG_INVALID,							item_offset(bmv.alarms.lowStarterVoltage),			IOPT( 0,		"Alarms/LowStarterVoltage",			&none,		0),	},
	{VE_REG_INVALID,							item_offset(bmv.alarms.highStarterVoltage),			IOPT( 0,		"Alarms/HighStarterVoltage",		&none,		0),	},
	{VE_REG_INVALID,							item_offset(bmv.alarms.lowSoc),						IOPT( 0,		"Alarms/LowSoc",					&none,		0),	},
	{VE_REG_INVALID,							item_offset(bmv.alarms.lowTemperature),				IOPT( 0,		"Alarms/LowTemperature",			&none,		0),	},
	{VE_REG_INVALID,							item_offset(bmv.alarms.highTemperature),			IOPT( 0,		"Alarms/HighTemperature",			&none,		0),	},
	{VE_REG_INVALID,							item_offset(bmv.alarms.midVoltage),					IOPT( 0,		"Alarms/MidVoltage",				&none,		0),	},
	{VE_REG_INVALID,							item_offset(bmv.settings.hasTemperature),			IOPT( 0,		"Settings/HasTemperature",			&none,		0),	},
	{VE_REG_INVALID,							item_offset(bmv.settings.hasStarterVoltage),		IOPT( 0,		"Settings/HasStarterVoltage",		&none,		0),	},
	{VE_REG_INVALID,							item_offset(bmv.settings.hasMidVoltage),			IOPT( 0,		"Settings/HasMidVoltage",			&none,		0),	},
#endif
	{VE_REG_DC_CHANNEL1_VOLTAGE,				item_offset(bmv.dcStatusBattery.voltage),			IOPT( DIV_V,	"Dc/0/Voltage",						&dcV,		0),	},
	{VE_REG_DC_CHANNEL1_CURRENT,				item_offset(bmv.dcStatusBattery.current),			IOPT( DIV_I,	"Dc/0/Current",						&dcI,		0),	},
	{VE_REG_DC_CHANNEL1_POWER,					item_offset(bmv.dcStatusBattery.power),				IOPT( DIV_WB,	"Dc/0/Power",						&dcW,		0),	},
	{VE_REG_BAT_TEMPERATURE,					item_offset(bmv.dcStatusBattery.temperature),		IOPT( DIV_T,	"Dc/0/Temperature",					&dcT,		0),	},
	{VE_REG_BATTERY_MID_POINT_VOLTAGE,			item_offset(bmv.dcStatusBattery.midVoltage),		IOPT( DIV_V,	"Dc/0/MidVoltage",					&dcV,		0),	},
	{VE_REG_BATTERY_MID_POINT_DEVIATION,		item_offset(bmv.dcStatusBattery.midVoltageDev),		IOPT( DIV_PCT,	"Dc/0/MidVoltageDeviation",			&percent,	0),	},
	{VE_REG_DC_CHANNEL2_VOLTAGE,				item_offset(bmv.dcStatusAux.voltage),				IOPT( DIV_V,	"Dc/1/Voltage",						&dcV,		0),	},
	{VE_REG_CAH,								item_offset(bmv.consumedAmphours),					IOPT( DIV_AH,	"ConsumedAmphours",					&ampH,		0),	},
	{VE_REG_SOC,								item_offset(bmv.stateOfCharge),						IOPT( DIV_SOC,	"Soc",								&percent,	0),	},
	{VE_REG_TTG,								item_offset(bmv.timeToGo),							IOPT( DIV_TTG,	"TimeToGo",							&sec,		0),	},
	{VE_REG_ALARM_REASON,						item_offset(bmv.alarms.reason),						IOPT( 0,		NULL,								&none,		0),	},
	{VE_REG_INVALID,							item_offset(bmv.alarms.flags),						IOPT( 0,		NULL,								&none,		0),	},
	{VE_REG_RELAY_CONTROL,						item_offset(bmv.alarms.relayState),					IOPT( 0,		"Relay/0/State",					&none,		0),	},
	{VE_REG_HIST_DEEPEST_DISCHARGE,				item_offset(bmv.history.deepestDischarge),			IOPT( DIV_AH,	"History/DeepestDischarge",			&ampH,		0),	},
	{VE_REG_HIST_LAST_DISCHARGE,				item_offset(bmv.history.lastDischarge),				IOPT( DIV_AH,	"History/LastDischarge",			&ampH,		0),	},
	{VE_REG_HIST_AVERAGE_DISCHARGE,				item_offset(bmv.history.averageDischarge),			IOPT( DIV_AH,	"History/AverageDischarge",			&ampH,		0),	},
	{VE_REG_HIST_NR_OF_CHARGE_CYCLES,			item_offset(bmv.history.chargeCycles),				IOPT( 0,		"History/ChargeCycles",				&none,		0),	},
	{VE_REG_HIST_NR_OF_FULL_DISCHARGES,			item_offset(bmv.history.fullDischarges),			IOPT( 0,		"History/FullDischarges",			&none,		0),	},
	{VE_REG_HIST_CUMULATIVE_AH,					item_offset(bmv.history.totalAhDrawn),				IOPT( DIV_AH,	"History/TotalAhDrawn",				&ampH,		0),	},
	{VE_REG_HIST_MIN_VOLTAGE,					item_offset(bmv.history.minimumVoltage),			IOPT( DIV_V,	"History/MinimumVoltage",			&dcV,		0),	},
	{VE_REG_HIST_MAX_VOLTAGE,					item_offset(bmv.history.maximumVoltage),			IOPT( DIV_V,	"History/MaximumVoltage",			&dcV,		0),	},
	{VE_REG_HIST_SECS_SINCE_LAST_FULL_CHARGE,	item_offset(bmv.history.timeSinceLastFullCharge),	IOPT( 0,		"History/TimeSinceLastFullCharge",	&sec,		0),	},
	{VE_REG_HIST_NR_OF_AUTO_SYNCS,				item_offset(bmv.history.automaticSyncs),			IOPT( 0,		"History/AutomaticSyncs",			&none,		0),	},
	{VE_REG_HIST_NR_OF_LOW_VOLTAGE_ALARMS,		item_offset(bmv.history.lowVoltageAlarms),			IOPT( 0,		"History/LowVoltageAlarms",			&none,		0),	},
	{VE_REG_HIST_NR_OF_HIGH_VOLTAGE_ALARMS,		item_offset(bmv.history.highVoltageAlarms),			IOPT( 0,		"History/HighVoltageAlarms",		&none,		0),	},
	{VE_REG_HIST_NR_OF_LOW_VOLTAGE_2_ALARMS,	item_offset(bmv.history.lowStarterVoltageAlarms),	IOPT( 0,		"History/LowStarterVoltageAlarms",	&none,		0),	},
	{VE_REG_HIST_NR_OF_HIGH_VOLTAGE_2_ALARMS,	item_offset(bmv.history.highStarterVoltageAlarms),	IOPT( 0,		"History/HighStarterVoltageAlarms",	&none,		0),	},
	{VE_REG_HIST_MIN_VOLTAGE_2,					item_offset(bmv.history.minimumStarterVoltage),		IOPT( DIV_V,	"History/MinimumStarterVoltage",	&dcV,		0),	},
	{VE_REG_HIST_MAX_VOLTAGE_2,					item_offset(bmv.history.maximumStarterVoltage),		IOPT( DIV_V,	"History/MaximumStarterVoltage",	&dcV,		0),	},
	{VE_REG_HIST_KWH_OUT,						item_offset(bmv.history.dischargedEnergy),			IOPT( DIV_KWH,	"History/DischargedEnergy",			&kwh,		0),	},
	{VE_REG_HIST_KWH_IN,						item_offset(bmv.history.chargedEnergy),				IOPT( DIV_KWH,	"History/ChargedEnergy",			&kwh,		0),	},
};

#ifdef ITEM_TREE
static AlarmDef const alarmDef[] =
{
	{item_offset(bmv.alarms.lowVoltage),			ALARM_LOW_VOLTAGE		},
	{item_offset(bmv.alarms.highVoltage),			ALARM_HIGH_VOLTAGE		},
	{item_offset(bmv.alarms.lowStarterVoltage),		ALARM_LOW_VOLTAGE2		},
	{item_offset(bmv.alarms.highStarterVoltage),	ALARM_HIGH_VOLTAGE2		},
	{item_offset(bmv.alarms.lowSoc),				ALARM_LOW_SOC			},
	{item_offset(bmv.alarms.lowTemperature),		ALARM_LOW_TEMPERATURE	},
	{item_offset(bmv.alarms.highTemperature),		ALARM_HIGH_TEMPERATURE	},
	{item_offset(bmv.alarms.midVoltage),			ALARM_MID_VOLTAGE		}
};
#endif

#ifdef ITEM_TREE

/* Callback to do some extra stuff when an item is set */
static void itemSetEvent(VeDirectService *instance, VeItem *item)
{
	VeVariant variant;

	if (item == &instance->items.bmv.dcStatusAux.voltage) {
		veItemOwnerSet(&instance->items.bmv.settings.hasStarterVoltage, veVariantUn32(&variant, 1));
		veItemOwnerSet(&instance->items.bmv.settings.hasTemperature, veVariantUn32(&variant, 0));
		veItemInvalidate(&instance->items.bmv.temperature);
		veItemOwnerSet(&instance->items.bmv.settings.hasMidVoltage, veVariantUn32(&variant, 0));
		veItemInvalidate(&instance->items.bmv.dcStatusBattery.midVoltage);
		veItemInvalidate(&instance->items.bmv.dcStatusBattery.midVoltageDev);
	}

	if (item == &instance->items.bmv.dcStatusBattery.temperature) {
		veItemOwnerSet(&instance->items.bmv.settings.hasTemperature, veVariantUn32(&variant, 1));
		veItemOwnerSet(&instance->items.bmv.settings.hasStarterVoltage, veVariantUn32(&variant, 0));
		veItemInvalidate(&instance->items.bmv.dcStatusAux.voltage);
		veItemOwnerSet(&instance->items.bmv.settings.hasMidVoltage, veVariantUn32(&variant, 0));
		veItemInvalidate(&instance->items.bmv.dcStatusBattery.midVoltage);
		veItemInvalidate(&instance->items.bmv.dcStatusBattery.midVoltageDev);
	}

	if (item == &instance->items.bmv.dcStatusBattery.midVoltage) {
		veItemOwnerSet(&instance->items.bmv.settings.hasMidVoltage, veVariantUn32(&variant, 1));
		veItemOwnerSet(&instance->items.bmv.settings.hasStarterVoltage, veVariantUn32(&variant, 0));
		veItemInvalidate(&instance->items.bmv.dcStatusAux.voltage);
		veItemOwnerSet(&instance->items.bmv.settings.hasTemperature, veVariantUn32(&variant, 0));
		veItemInvalidate(&instance->items.bmv.temperature);
	}
}

static char const *batteryName(void)
{
	 return "com.victronenergy.battery";
}

#endif

static void itemUpdate (VeDirectService * const instance, const VetiId id, const VeVariantValue value, const VetiTypes type, const veBool valid)
{
	unsigned int i;
	VeVariant v, v2;
	const ItemInfo * item = NULL;

	logI(MODULE, "itemUpdate %4.4X %8.8X %d", id, value.UN32, valid);

	/* First handle the special cases */
	/* split up the alarm reason to separate item paths */
#ifdef ITEM_TREE
	if (id == VE_REG_ALARM_REASON && valid) {
		for (i = 0; i < ARRAY_LENGTH(alarmDef); i++) {
			veVariantUn32(&v, value.UN16 & alarmDef[i].mask ? DBUS_ALARM_ERROR : DBUS_ALARM_NONE);
			veItemOwnerSet(instance_ptr(instance, alarmDef[i].item), &v);
		}
	}
#endif

	if (valid
			&& (id == VE_REG_DC_CHANNEL1_CURRENT || id == VE_REG_DC_CHANNEL1_VOLTAGE)
			&& vetiGetValue(&instance->veti, VE_REG_DC_CHANNEL1_CURRENT, &v.value)
			&& vetiGetValue(&instance->veti, VE_REG_DC_CHANNEL1_VOLTAGE, &v2.value)
			&& !vetiIsValid(&instance->veti, VE_REG_DC_CHANNEL1_POWER)) {
		veItemOwnerSet(&instance->items.bmv.dcStatusBattery.power, veVariantFloat(&v, (sn32) v.value.SN16 * v2.value.UN16 / (DIV_I * DIV_V)));
	}

	for (i = 0; i < ARRAY_LENGTH(itemInfo) && !item; ++i) {
		if (itemInfo[i].veReg == id) {
			item = &itemInfo[i];
		}
	}

	/* TODO: alarm.flags when compiling this for CAN cable */

	if (!item)
		return;

	if (item->id)
		logI(MODULE, "Found item %s", item->id);

	switch (type) {
	case VETI_UN8:
		veVariantToN32(veVariantUn8(&v, value.UN8));
		break;
	case VETI_SN8:
		veVariantToN32(veVariantSn8(&v, value.SN8));
		break;
	case VETI_UN16:
		veVariantToN32(veVariantUn16(&v, value.UN16));
		break;
	case VETI_SN16:
		veVariantToN32(veVariantSn16(&v, value.SN16));
		break;
	case VETI_UN32:
		veVariantToN32(veVariantUn32(&v, value.UN32));
		break;
	case VETI_SN32:
		veVariantToN32(veVariantSn32(&v, value.SN32));
		break;
	/* We don't know the size of the item, so really nothing we can do but handle
	 * it above with the specific items. */
	case VETI_PTR:
		logE(MODULE, "Received VETI_PTR item update for %04X. We can not handle this in a generic way. Exiting...", id);
		pltExit(146);

	case VETI_UNKNOWN:
		logE(MODULE, "Received VETI_UKNOWN item update for %04X. We can not handle this in a generic way. Exiting...", id);
		pltExit(147);
	}

#ifdef ITEM_TREE
	if (item->veReg == VE_REG_BAT_TEMPERATURE && veVariantIsValid(&v)) {
		veVariantSn32(&v, (sn32) v.value.UN32 - 27315);
	}
#endif

	if (valid)
		itemsSetItem(instance_ptr(instance, item->item_offset), &v, item);
	else
		veItemInvalidate(instance_ptr(instance, item->item_offset));

	itemSetEvent(instance, instance_ptr(instance, item->item_offset));
}

static veBool hasSerialNumber(un16 pid)
{
	switch (pid)
	{
	case VE_PROD_ID_BMV600S:
	case VE_PROD_ID_BMV600HS:
	case VE_PROD_ID_BMV602S:
		// Can only be retrieved using hex and we cannot talk hex to bmv60x
		return veFalse;
	default:
		return veTrue;
	}
}

static void connectedEvent(VeDirectService * const instance, un16 pid)
{
	VeVariant variant;
	un16 i;

	logI(MODULE, "found %s", veProductGetName(pid));

	/* Build tree and bus on */
	for (i = 0; i < ARRAY_LENGTH(itemInfo); i++)
		itemsAddItemByInfo(&instance->root, instance_ptr(instance, itemInfo[i].item_offset), &itemInfo[i], NULL, NULL);

	veItemOwnerSet(&instance->items.bmv.settings.hasStarterVoltage, veVariantUn32(&variant, 0));
	veItemOwnerSet(&instance->items.bmv.settings.hasMidVoltage, veVariantUn32(&variant, 0));
	veItemOwnerSet(&instance->items.bmv.settings.hasTemperature, veVariantUn32(&variant, 0));

	/* Retrieve the serial number for the unique id and the custom name */
	if (hasSerialNumber(pid)) {
		vetiPublishRequest(&instance->veti, instance->vetiHandle, VE_REG_SERIAL_NUMBER);
	}
}

static const VeDirectProduct veDirectBmv = {
	NULL,
	NULL,
	NULL,
	itemUpdate,
	hasSerialNumber,

	connectedEvent,
	NULL,

#ifdef ITEM_TREE
	batteryName,
#endif

};

const VeDirectProduct *veDirectBmvFunc(un16 pid)
{
	/* Follow the classification of veti */
	if (vdVetiClass(pid) == classBattery)
		return &veDirectBmv;

	return NULL;
}
