#include <velib/utils/ve_logger.h>
#include <velib/vecan/products.h>
#include <velib/platform/console.h>

#include "vedirect_service.h"
#include "items_helpers.h"

/* TODO:
 * How to handle alarms / warnings / DBUS_ALARM_ERROR / DBUS_ALARM_WARNING?
 *
 * */

#define MODULE	"inverter"

#define ALARM_LOW_VOLTAGE			0x0001
#define ALARM_HIGH_VOLTAGE			0x0002
#define ALARM_LOW_SOC				0x0004
#define ALARM_LOW_VOLTAGE2			0x0008
#define ALARM_HIGH_VOLTAGE2			0x0010
#define ALARM_LOW_TEMPERATURE		0x0020
#define ALARM_HIGH_TEMPERATURE		0x0040
#define ALARM_MID_VOLTAGE			0x0080
#define ALARM_OVERLOAD				0x0100
#define ALARM_DC_RIPPLE				0x0200
#define ALARM_LOW_VOLTAGE_AC_OUT	0x0400
#define ALARM_HIGH_VOLTAGE_AC_OUT	0x0800

typedef enum
{
	DBUS_ALARM_NONE,
	DBUS_ALARM_WARNNG,
	DBUS_ALARM_ERROR
} DBusAlarm;

typedef struct
{
	size_t item;
	un16 mask;
} AlarmDef;

static ItemInfo const itemInfo[] =
{
#ifdef ITEM_TREE
	{VE_REG_INVALID,							item_offset(inverter.alarms.lowVoltage),				IOPT( 0,		"Alarms/LowVoltage",			&none,		veFalse ), },
	{VE_REG_INVALID,							item_offset(inverter.alarms.highVoltage),				IOPT( 0,		"Alarms/HighVoltage",			&none,		veFalse ), },
	{VE_REG_INVALID,							item_offset(inverter.alarms.lowTemperature),			IOPT( 0,		"Alarms/LowTemperature",		&none,		veFalse ), },
	{VE_REG_INVALID,							item_offset(inverter.alarms.highTemperature),			IOPT( 0,		"Alarms/HighTemperature",		&none,		veFalse ), },
	{VE_REG_INVALID,							item_offset(inverter.alarms.overload),					IOPT( 0,		"Alarms/Overload",				&none,		veFalse ), },
	{VE_REG_INVALID,							item_offset(inverter.alarms.dcRipple),					IOPT( 0,		"Alarms/Ripple",				&none,		veFalse ), },
	{VE_REG_INVALID,							item_offset(inverter.alarms.lowVoltageAcOut),			IOPT( 0,		"Alarms/LowVoltageAcOut",		&none,		veFalse ), },
	{VE_REG_INVALID,							item_offset(inverter.alarms.highVoltageAcOut),			IOPT( 0,		"Alarms/HighVoltageAcOut",		&none,		veFalse ), },
#endif
	{VE_REG_DC_CHANNEL1_VOLTAGE,				item_offset(inverter.dcBatteryStatus.voltage),			IOPT( DIV_V,	"Dc/0/Voltage",					&dcV,		veFalse ), },
	{VE_REG_WARNING_REASON,						item_offset(inverter.warningReason),					IOPT( 0,		NULL,							&none,		veFalse ), },
	{VE_REG_ALARM_REASON,						item_offset(inverter.alarmReason),						IOPT( 0,		NULL,							&none,		veFalse ), },
	{VE_REG_AC_OUT_VOLTAGE,						item_offset(inverter.acOutStatus.voltage),				IOPT( DIV_V,	"Ac/Out/L1/V",					&acV,		veFalse ), },
	{VE_REG_AC_OUT_CURRENT,						item_offset(inverter.acOutStatus.current),				IOPT( DIV_I,	"Ac/Out/L1/I",					&acI,		veFalse ), },
	{VE_REG_DEVICE_MODE,						item_offset(inverter.deviceMode),						IOPT( 0,		"Mode",							&none,		veFalse ), },
	{VE_REG_DEVICE_STATE,						item_offset(inverter.deviceState),						IOPT( 0,		"State",						&none,		veFalse ), },
};

#ifdef ITEM_TREE
static AlarmDef const alarmDef[] =
{
	{item_offset(inverter.alarms.lowVoltage),			ALARM_LOW_VOLTAGE			},
	{item_offset(inverter.alarms.highVoltage),			ALARM_HIGH_VOLTAGE			},
	{item_offset(inverter.alarms.lowTemperature),		ALARM_LOW_TEMPERATURE		},
	{item_offset(inverter.alarms.highTemperature),		ALARM_HIGH_TEMPERATURE		},
	{item_offset(inverter.alarms.overload),				ALARM_OVERLOAD				},
	{item_offset(inverter.alarms.dcRipple),				ALARM_DC_RIPPLE				},
	{item_offset(inverter.alarms.lowVoltageAcOut),		ALARM_LOW_VOLTAGE_AC_OUT	},
	{item_offset(inverter.alarms.highVoltageAcOut),		ALARM_HIGH_VOLTAGE_AC_OUT	},
};
#endif

static void updateAlarms (VeDirectService *instance)
{
	int i;
	VeVariant v;
	un16 alarmReason = veItemLocalValue(&instance->items.inverter.alarmReason, &v)->value.UN16;
	un16 warningReason = veItemLocalValue(&instance->items.inverter.warningReason, &v)->value.UN16;

	for (i = 0; i < ARRAY_LENGTH(alarmDef); i++) {
		veVariantUn32(&v, alarmReason & alarmDef[i].mask ? DBUS_ALARM_ERROR : warningReason & alarmDef[i].mask ? DBUS_ALARM_WARNNG : DBUS_ALARM_NONE);
		veItemOwnerSet(instance_ptr(instance, alarmDef[i].item), &v);
	}
}

static void itemUpdate (VeDirectService *instance, const VetiId id, VeVariantValue value, const VetiTypes type, const veBool valid)
{
	unsigned int i;
	const ItemInfo * item = NULL;
	VeVariant v;

	logI(MODULE, "itemUpdate for id %4.4x -> %8.8x (%d)", id, value.UN32, valid);

	/* First handle the special cases */
	/* split up the alarm and warning reason to separate item paths */
#ifdef ITEM_TREE
	if (id == VE_REG_ALARM_REASON && valid) {
		veVariantToN32(veVariantUn16(&v, value.UN16));
		veItemOwnerSet(&instance->items.inverter.alarmReason, &v);

		updateAlarms(instance);
	}
	if (id == VE_REG_WARNING_REASON && valid) {
		veVariantToN32(veVariantUn16(&v, value.UN16));
		veItemOwnerSet(&instance->items.inverter.warningReason, &v);

		updateAlarms(instance);
	}
#endif

	/* Standard items */
	for (i = 0; i < ARRAY_LENGTH(itemInfo) && !item; ++i) {
		if (itemInfo[i].veReg == id) {
			item = &itemInfo[i];
		}
	}

	if (!item)
		return;

	switch (type) {
	case VETI_UN8:
		veVariantToN32(veVariantUn8(&v, value.UN8));
		break;
	case VETI_SN8:
		veVariantToN32(veVariantSn8(&v, value.SN8));
		break;
	case VETI_UN16:
		veVariantToN32(veVariantUn16(&v, value.UN16));
		break;
	case VETI_SN16:
		veVariantToN32(veVariantSn16(&v, value.SN16));
		break;
	case VETI_UN32:
		veVariantToN32(veVariantUn32(&v, value.UN32));
		break;
	case VETI_SN32:
		veVariantToN32(veVariantSn32(&v, value.SN32));
		break;
	/* We don't know the size of the item, so really nothing we can do but handle
	 * it above with the specific items. */
	case VETI_PTR:
		logE(MODULE, "Received VETI_PTR item update for %04X. We can not handle this in a generic way. Exiting...", id);
		pltExit(146);

	case VETI_UNKNOWN:
		logE(MODULE, "Received VETI_UKNOWN item update for %04X. We can not handle this in a generic way. Exiting...", id);
		pltExit(147);
	}

	if (valid) {
		itemsSetItem(instance_ptr(instance, item->item_offset), &v, item);
	} else {
		veItemInvalidate(instance_ptr(instance, item->item_offset));
	}
}

static void connectedEvent(VeDirectService * const instance, un16 pid)
{
	un16 i;

	logI(MODULE, "found %s", veProductGetName(pid));

	/* Build tree and bus on */
	for (i = 0; i < ARRAY_LENGTH(itemInfo); i++)
		itemsAddItemByInfo(&instance->root, instance_ptr(instance, itemInfo[i].item_offset), &itemInfo[i], NULL, NULL);

}

#ifdef ITEM_TREE
static char const *inverterName(void)
{
	return "com.victronenergy.inverter";
}
#endif

static veBool hasSerialNumber(un16 pid)
{
	VE_UNUSED(pid);

	return veTrue;
}

static const VeDirectProduct veDirectInverter = {
	NULL,
	NULL, // tick,
	NULL,
	itemUpdate,
	hasSerialNumber,

	connectedEvent,
	NULL,

#ifdef ITEM_TREE
	inverterName,
#endif
};

const VeDirectProduct *veDirectInverterFunc(un16 pid)
{
	/* Follow the classification of veti */
	if (vdVetiClass(pid) == classInverter)
		return &veDirectInverter;

	return NULL;
}
