#include "vedirect_service.h"

#include <string.h>

#include <velib/utils/ve_logger.h>
#include <velib/base/ve_string.h>
#include <velib/platform/console.h>

#include "platform.h"
#include "vedirect_products.h"
#include "vedirect_formatters.h"
#include "items_helpers.h"
#include "serial_hal.h"
#include "values.h"

/* This file should support multiple instances of the vedirect_service and should
 * thus NOT contain any non-const global variables */

#define MODULE "service"

static const ItemInfo itemInfo[] =
{
	{VE_REG_INVALID,		item_offset(prodName),			IOPT( 0,	"ProductName",		&none,		veFalse ),	},
	{VE_REG_INVALID,		item_offset(hardwareVersion),	IOPT( 0,	"HardwareVersion",	&none,		veFalse ),	},
	{VE_REG_INVALID,		item_offset(instance),			IOPT( 0,	"DeviceInstance",	&none,		veFalse ),	},
	{VE_REG_INVALID,		item_offset(connected),			IOPT( 0,	"Connected",		&none,		veFalse ),	},
	{VE_REG_INVALID,		item_offset(prodId),			IOPT( 0,	"ProductId",		&none,		veFalse ),	},
	{VE_REG_APP_VER,		item_offset(firmwareVersion),	IOPT( 0,	"FirmwareVersion",	&none,		veFalse ),	},
	{VE_REG_SERIAL_NUMBER,	item_offset(serial),			IOPT( 0,	"Serial",			&none,		veFalse ),	},
};

static const char *unknownServiceName(void) {
	return "com.victronenergy.unknown";
}

static const VeDirectProduct unknownProduct = {
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
#ifdef ITEM_TREE
	unknownServiceName,
#endif
};
const VeDirectProduct *veDirectUnknownFunc(un16 pid) {
	VE_UNUSED(pid);

	return &unknownProduct;
}

static void disconnectedDevice(VeDirectService *instance)
{
	VeVariant variant;

	veItemOwnerSet(&instance->items.connected, veVariantUn32(&variant, 0));

	if (instance->veDirectProduct->disconnectedEvent)
		instance->veDirectProduct->disconnectedEvent(instance);
	if (instance->veDirectProduct->invalidate)
		instance->veDirectProduct->invalidate(instance);

#if 0
	veItemDeleteBranch(&instance->root);
	instance->registered = veFalse;
	instance->veDirectProduct = NULL;
#else
	pltExit(130);
#endif
}

static void registerDevice(VeDirectService *instance, un32 instanceNr, const char *servicePostfix)
{
	VeVariant v;

	veItemOwnerSet(instance_ptr(instance, item_offset(instance)), veVariantUn32(&v, instanceNr));

	ve_snprintf(instance->serviceName, sizeof(instance->serviceName), "%s.%s", instance->veDirectProduct->getServiceName(), servicePostfix);

	veItemInit(&instance->root, instance->serviceName, instance->rootParent);
	instance->registered = veTrue;
}

static void connectedDevice(VeDirectService *instance, un16 pid)
{
	unsigned int i;
	VeVariant variant;
	VeDirectFunc const *func = productsFuncs();

	/* Build item tree*/
	for (i = 0; i < ARRAY_LENGTH(itemInfo); i++) {
		VeItem *item = instance_ptr(instance, itemInfo[i].item_offset);
		veItemCtx(item)->ptr = instance;
		itemsAddItemByInfo(&instance->root, item, &itemInfo[i], NULL, NULL);
	}

#ifdef ITEM_TREE
	/* Set formatters */
	veItemSetFmt(instance_ptr(instance, item_offset(prodId)), pidFmt, NULL);
	veItemSetFmt(instance_ptr(instance, item_offset(firmwareVersion)), versionFmt, NULL);
#endif

	veItemOwnerSet(instance_ptr(instance, item_offset(prodId)), veVariantUn32(&variant, pid));
	veItemOwnerSet(instance_ptr(instance, item_offset(prodName)), veVariantStr(&variant, veProductGetName(pid)));
	veItemOwnerSet(instance_ptr(instance, item_offset(connected)), veVariantUn32(&variant, 1));

	while (*func) {
		const VeDirectProduct *prod = (*func)(pid);
		if (prod) {
			instance->prodId = pid;
			instance->veDirectProduct = prod;

			if (instance->veDirectProduct->connectedEvent)
				instance->veDirectProduct->connectedEvent(instance, pid);

			registerDevice(instance, instance->defaultInstanceId, instance->defaultServicePostfix);

			return;
		}
		func++;
	}

	logW(MODULE, "found unsupported VE.Direct device %4.4X", pid);
}

static void statusChange(const VdDirectStatus status, VeDirectService * const instance)
{
	vdVetiStatusChange(status, &instance->vdVeti);

	if (status == VdIdentified) {
		/* when we are already connected and get a new Identified, it means
		 * that the found PID has changed, so report the previous device
		 * disconnected. */
		if (instance->connected) {
			logE(MODULE, "New connected product identified");
			disconnectedDevice(instance);
			instance->connected = veFalse;
		}

		connectedDevice(instance, vdGetProdId(&instance->vd));

		instance->connected = veTrue;
	} else {
		if (instance->connected) {
			logE(MODULE, "connection lost");
			disconnectedDevice(instance);
			instance->connected = veFalse;
		}
	}
}

static void txData(const un8 * data, const un8 length, VeDirectService * const instance)
{
	/* length is a un8, so max 255, so a buffer of 255+1 (trailing zero) is enough */
	un8 buf[256];

	memcpy(buf, data, length);
	buf[length] = 0;

	logI(MODULE, "%p: sending %d %s", instance, length, buf);

	if (instance->txFunc)
		instance->txFunc(data, length, instance->txContext);
}

static void genericItemUpdate(VeDirectService * const instance, const VetiId id, const VeVariantValue value, const VetiTypes type, const veBool valid)
{
	ItemInfo const * item = NULL;
	unsigned int i;

	for (i = 0; i < ARRAY_LENGTH(itemInfo) && !item; ++i) {
		if (itemInfo[i].veReg == id) {
			item = &itemInfo[i];
		}
	}

	if (!item)
		return;

	if (item->id)
		logI(MODULE, "Found item %s", item->id);

	if (type == VETI_PTR && valid && value.Ptr != NULL) {
		/* Here all the items that are VETI_PTR types should be handled */
		if (item->item_offset == item_offset(serial)) {
			VeVariant v;
			char str[12];

			/* Check if we already received a serial number and it is different.
			 * This can happen when the user switches connections to MPPTs of the same type
			 * really fast so a disconnection is not triggered. */
			if (instance->ser[0] && memcmp(instance->ser, value.CPtr, 11) != 0) {
				disconnectedDevice(instance);
				return;
			}

			/* create a zero terminated string */
			memcpy(str, value.CPtr, 11);
			str[11] = 0;

			/* Store the serial number in the VeItem */
			itemsSetBufferPtr(&instance->items.serial, veVariantStr(&v, str), instance->ser, 12);

			return;
		}
	}

	if (valid) {
		VeVariant v;

		switch (item->veReg) {
		case VE_REG_APP_VER:
			instance->appVersion = value.UN16;
			break;
		default:
			break;
		}

		switch (type) {
		case VETI_UN8:
			veVariantToN32(veVariantUn8(&v, value.UN8));
			break;
		case VETI_SN8:
			veVariantToN32(veVariantSn8(&v, value.SN8));
			break;
		case VETI_UN16:
			veVariantToN32(veVariantUn16(&v, value.UN16));
			break;
		case VETI_SN16:
			veVariantToN32(veVariantSn16(&v, value.SN16));
			break;
		case VETI_UN32:
			veVariantToN32(veVariantUn32(&v, value.UN32));
			break;
		case VETI_SN32:
			veVariantToN32(veVariantSn32(&v, value.SN32));
			break;
		default:
			/* All pointer types should be handled above or are invalid */
			v.value.CPtr = NULL;
			v.type.tp = VE_BUF;
		}

		itemsSetItem(instance_ptr(instance, item->item_offset), &v, item);
	} else {
		veItemInvalidate(instance_ptr(instance, item->item_offset));
	}
}

static void itemUpdate(void * const context, const VetiId id, const VeVariantValue * value, const veBool valid)
{
	VeDirectService * const instance = context;
	VetiTypes type = vetiGetType(&instance->veti, id);

	genericItemUpdate(instance, id, *value, type, valid);

	if (instance->veDirectProduct && instance->veDirectProduct->itemUpdate)
		instance->veDirectProduct->itemUpdate(instance, id, *value, type, valid);
}

static void textItemNotify (const VeRegId id, const VeVariantValue * value, const VeDataBasicType type, VeDirectService * instance)
{
	vdVetiTextItemNotify(id, value, type, &instance->vdVeti);
}

static void hexMessageNotify (const VeRegId id, const un8 * const data, const un8 length, const un8 flags, veBool async, VeDirectService * const instance)
{
	vdVetiHexMessageNotify(id, data, length, flags, async, &instance->vdVeti);
}

static void hexErrorNotify(const VdHexError error, VeDirectService * const instance)
{
	vdVetiHexErrorNotify(error, &instance->vdVeti);

	if (error == VdHexBusy) {
		logE(MODULE, "Received a hexBusy. This is an error in the application. Exiting...");
		pltExit(134);
	} else if (error == VdHexSetTypeUnsupported) {
		logE(MODULE, "Received a VdHexSetTypeUnsupported. Exiting...");
		pltExit(135);
	} else if (error == VdHexTxBufferOverflow) {
		logE(MODULE, "Received a VdHexTxBufferOverflow. Exiting...");
		pltExit(136);
	}
}

static void vdVetiErrorNotify(const VdVetiError error, void * const context)
{
	VE_UNUSED(context);
	logE(MODULE, "Received a vdVetiError %d. Exiting...", error);
	pltExit(140);
}

void veDirectServiceInit(VeDirectService * const instance, VeItem *rootParent, un32 defaultInstanceId, const char *defaultServicePostfix)
{
	memset(instance, 0, sizeof(*instance));

	vdInit(&instance->vd, NULL);
	vetiInit(&instance->veti);
	vdVetiInit(&instance->vdVeti, &instance->vd, &instance->veti, instance);
	vdSetContext(&instance->vd, instance);

	vdConnect(&instance->vd, TxData, txData);
	vdConnect(&instance->vd, StatusChange, statusChange);
	vdConnect(&instance->vd, TextItemNotify, textItemNotify);
	vdConnect(&instance->vd, HexMessageNotify, hexMessageNotify);
	vdConnect(&instance->vd, HexErrorNotify, hexErrorNotify);

	vdVetiConnectErrorNotify(&instance->vdVeti, vdVetiErrorNotify);

	instance->vetiHandle = vetiConnect(&instance->veti);
	if (!vetiConnectSubscriber(&instance->veti, instance->vetiHandle, itemUpdate, instance)) {
		logE(MODULE, "Connecting veti subscriber failed. Exiting...");
		pltExit(133);
	}
	vetiSubscribeAll(&instance->veti, instance->vetiHandle);

	instance->rootParent = rootParent;
	instance->defaultInstanceId = defaultInstanceId;
	instance->defaultServicePostfix = defaultServicePostfix;
	instance->prodId = VE_PROD_NOT_SET;
	instance->appVersion = 0xFFFF;
}

void veDirectServiceTick(VeDirectService *instance)
{
	static const un16 stackTickTime = 100;

	instance->vdTickCnt += TICK_TIME;
	if (instance->vdTickCnt >= stackTickTime) {
		instance->vdTickCnt -= stackTickTime;
		vdTick(&instance->vd);
	}

	if (instance->veDirectProduct && instance->veDirectProduct->tick)
		instance->veDirectProduct->tick(instance);
}

