#include <string.h>

#include <velib/base/ve_string.h>
#include <velib/types/types.h>
#include <velib/nmea2k/n2k.h>		// Converter state
#include <velib/vecan/charger_error.h>
#include <velib/utils/ve_logger.h>
#include <velib/platform/console.h>

#include "platform.h"
#include "values.h"
#include "items_helpers.h"
#include "vedirect_service.h"

#define ASSERT_CONCAT_(a, b) a##b
#define ASSERT_CONCAT(a, b) ASSERT_CONCAT_(a, b)
#define STATIC_ASSERT(e,m) \
	typedef char ASSERT_CONCAT(assert_array_line_, __LINE__)[(e) ? 1 : -1]

/* This file should support multiple instances of the vedirect_service and should
 * thus NOT contain any non-const global variables */

#define MODULE	"solar"

static const ItemInfo itemInfo[] =
{
	{VE_REG_DC_CHANNEL1_VOLTAGE,			item_offset(solar.dcStatusBattery.voltage),				IOPT( DIV_V,	"Dc/0/Voltage",							&dcV,		veFalse ), },
	{VE_REG_DC_CHANNEL1_CURRENT,			item_offset(solar.dcStatusBattery.current),				IOPT( DIV_I,	"Dc/0/Current",							&dcI,		veFalse ), },
	{VE_REG_DC_INPUT_VOLTAGE,				item_offset(solar.dcStatusPvPanel.voltage),				IOPT( DIV_V,	"Pv/V",									&dcV,		veFalse ), },
	{VE_REG_DC_INPUT_CURRENT,				item_offset(solar.dcStatusPvPanel.current),				IOPT( DIV_I,	"Pv/I",									&dcI,		veFalse ), },
	{VE_REG_DC_INPUT_POWER,					item_offset(solar.dcStatusPvPanel.power),				IOPT( DIV_WPV,	"Yield/Power",							&dcW,		veFalse ), },
	{VE_REG_CHR_ERROR_CODE,					item_offset(solar.error),								IOPT( 0,		"ErrorCode",							&none,		veFalse ), },
	{VE_REG_DEVICE_STATE,					item_offset(solar.state),								IOPT( 0,		"State",								&none,		veFalse ), },
	{VE_REG_DC_OUTPUT_STATUS,				item_offset(solar.load.state),							IOPT( 0,		"Load/State",							&none,		veTrue ), },
	{VE_REG_DC_OUTPUT_CURRENT,				item_offset(solar.load.current),						IOPT( DIV_I,	"Load/I",								&dcI,		veTrue ), },
	{VE_REG_RELAY_CONTROL,					item_offset(solar.relayState),							IOPT( 0,		"Relay/0/State",						&none,		veTrue ),	},

	/* Statistic values. */
	{VE_REG_CHR_USER_YIELD,					item_offset(solar.histLast.yieldUser),					IOPT( DIV_KWH,	"Yield/User",							&kwh,		veFalse ), },

	/* Daily history values. */
	{VE_REG_CHR_TODAY_YIELD,				item_offset(solar.histDay[0].yield),					IOPT( DIV_KWH,	"History/Daily/0/Yield",				&none,		veFalse ), },
	{VE_REG_CHR_TODAY_PMAX,					item_offset(solar.histDay[0].maxPower),					IOPT( 1,		"History/Daily/0/MaxPower",				&none,		veFalse ), },
	{VE_REG_CHR_YESTERDAY_YIELD,			item_offset(solar.histDay[1].yield),					IOPT( DIV_KWH,	"History/Daily/1/Yield",				&none,		veFalse ), },
	{VE_REG_CHR_YESTERDAY_PMAX,				item_offset(solar.histDay[1].maxPower),					IOPT( 1,		"History/Daily/1/MaxPower",				&none,		veFalse ), },

	{VE_REG_BAT_MAX_CURRENT,				item_offset(solar.settings.chargeCurrentLimit),			IOPT( DIV_I,	"Settings/ChargeCurrentLimit",			&dcI,		veFalse ), },
	{VE_REG_BAT_BMS_PRESENT,				item_offset(solar.settings.bmsPresent),					IOPT( 0,		"Settings/BmsPresent",					&none,		veTrue ), },
	{VE_REG_LINK_VSET,						item_offset(solar.link.chargeVoltageSet),				IOPT( DIV_V,	"Link/ChargeVoltage",					&dcV,		veTrue ), },
	{VE_REG_LINK_NETWORK_MODE,				item_offset(solar.link.networkMode),					IOPT( 0,		"Link/NetworkMode",						&none,		veTrue ), },
	{VE_REG_LINK_NETWORK_STATUS,			item_offset(solar.link.networkStatus),					IOPT( 0,		"Link/NetworkStatus",					&none,		veTrue ), },
	{VE_REG_LINK_CHARGE_CURRENT_PERCENTAGE,	item_offset(solar.link.chargeCurrentSet),				IOPT( DIV_I,	"Link/ChargeCurrent",					&dcI,		veTrue ), },
	{VE_REG_LINK_CHARGE_CURRENT_LIMIT,		item_offset(solar.link.chargeCurrentSet),				IOPT( DIV_I,	"Link/ChargeCurrent",					&dcI,		veTrue ), },
	{VE_REG_CAPABILITIES1,					item_offset(solar.capabilities.capabilities1),			IOPT( 0,		"Capabilities/Capabilities1",			&none,		veFalse ), },
};

/* These items are requested on connection. This should be items not received via text,
 * so typically settings. */
static const VeRegId requestItems[] =
{
	VE_REG_BAT_MAX_CURRENT,
	VE_REG_CAPABILITIES1,
};

typedef enum {
	totalFieldFirst,
	totalYieldSystem = totalFieldFirst,
	totalYieldUser,
	totalMaxPvVoltage,
	totalMaxBatteryVoltage,
	totalMinBatteryVoltage,
	totalLastError0,
	totalLastError1,
	totalLastError2,
	totalLastError3,
	totalDaysAvailable,
	totalRaw,
	totalFieldLast = totalRaw,
} HistoryTotalFields;

/* This order should be the same as HistoryTotalFields */
static const ItemInfo historyTotalItemInfo[] = {
	{VE_REG_INVALID,	item_offset(solar.histLast.yieldSystem),		IOPT( DIV_KWH,	"Yield/System",							&kwh,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histLast.yieldUser),			IOPT( DIV_KWH,	"Yield/User",							&kwh,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histLast.maxPvVoltage),		IOPT( DIV_V,	"History/Overall/MaxPvVoltage",			&dcV,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histLast.maxBatteryVoltage),	IOPT( DIV_V,	"History/Overall/MaxBatteryVoltage",	&dcV,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histLast.minBatteryVoltage),	IOPT( DIV_V,	"History/Overall/MinBatteryVoltage",	&dcV,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histLast.lastError[0]),		IOPT( 0,		"History/Overall/LastError1",			&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histLast.lastError[1]),		IOPT( 0,		"History/Overall/LastError2",			&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histLast.lastError[2]),		IOPT( 0,		"History/Overall/LastError3",			&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histLast.lastError[3]),		IOPT( 0,		"History/Overall/LastError4",			&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histLast.daysAvailable),		IOPT( 0,		"History/Overall/DaysAvailable",		&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histLast.raw),				IOPT( 0,		"History/Overall/Raw",					&none,	veFalse ), },
};
STATIC_ASSERT((totalFieldLast - totalFieldFirst + 1) == ARRAY_LENGTH(historyTotalItemInfo), "historyTotalItemInfo should match HistoryTotalFields");

typedef enum {
	dailyFieldFirst,
	dailySeqNr = dailyFieldFirst,
	dailyYield,
	dailyConsumption,
	dailyMaxBatteryVoltage,
	dailyMinBatteryVoltage,
	dailyMaxCurrent,
	dailyMaxPower,
	dailyMaxPvVoltage,
	dailyLastError0,
	dailyLastError1,
	dailyLastError2,
	dailyLastError3,
	dailyBulkTime,
	dailyAbsTime,
	dailyFloatTime,
	dailyRaw,
	dailyFieldLast = dailyRaw,
} HistoryDailyFields;

static const ItemInfo historyDailyItemInfo[] = {
	{VE_REG_INVALID,	item_offset(solar.histDay[0].seqNr),				IOPT( 1,		"History/Daily/%d/Nr",					&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histDay[0].yield),				IOPT( DIV_KWH,	"History/Daily/%d/Yield",				&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histDay[0].consumption),			IOPT( DIV_KWH,	"History/Daily/%d/Consumption",			&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histDay[0].maxBatteryVoltage),	IOPT( DIV_V,	"History/Daily/%d/MaxBatteryVoltage",	&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histDay[0].minBatteryVoltage),	IOPT( DIV_V,	"History/Daily/%d/MinBatteryVoltage",	&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histDay[0].maxCurrent),			IOPT( DIV_I,	"History/Daily/%d/MaxBatteryCurrent",	&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histDay[0].maxPower),				IOPT( 1,		"History/Daily/%d/MaxPower",			&none,	veFalse ), }, /* Total history, power is in W and not 0.01W */
	{VE_REG_INVALID,	item_offset(solar.histDay[0].maxPvVoltage),			IOPT( DIV_V,	"History/Daily/%d/MaxPvVoltage",		&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histDay[0].lastError[0]),			IOPT( 0,		"History/Daily/%d/LastError1",			&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histDay[0].lastError[1]),			IOPT( 0,		"History/Daily/%d/LastError2",			&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histDay[0].lastError[2]),			IOPT( 0,		"History/Daily/%d/LastError3",			&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histDay[0].lastError[3]),			IOPT( 0,		"History/Daily/%d/LastError4",			&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histDay[0].bulkTime),				IOPT( 1,		"History/Daily/%d/TimeInBulk",			&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histDay[0].absTime),				IOPT( 1,		"History/Daily/%d/TimeInAbsorption",	&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histDay[0].floatTime),			IOPT( 1,		"History/Daily/%d/TimeInFloat",			&none,	veFalse ), },
	{VE_REG_INVALID,	item_offset(solar.histDay[0].raw),					IOPT( 0,		"History/Daily/%d/Raw",					&none,	veFalse ), },
};
STATIC_ASSERT((dailyFieldLast - dailyFieldFirst + 1) == ARRAY_LENGTH(historyDailyItemInfo), "historyDailyItemInfo should match HistoryDailyFields");

#ifdef ITEM_TREE

static char const *solarName(void)
{
	return "com.victronenergy.solarcharger";
}

#endif

static const ItemInfo * getItemInfoFromId(VeRegId id)
{
	unsigned int i;
	for (i = 0; i < ARRAY_LENGTH(itemInfo); ++i) {
		if (itemInfo[i].veReg == id) {
			return &itemInfo[i];
		}
	}

	return NULL;
}

static const ItemInfo * getItemInfoFromItem(const VeDirectService * instance, const VeItem * item)
{
	unsigned int i;
	for (i = 0; i < ARRAY_LENGTH(itemInfo); i++) {
		if (instance_ptr(instance, itemInfo[i].item_offset) == item) {
			return &itemInfo[i];
		}
	}

	return NULL;
}

static void setVetiRequest(VeDirectService * instance, VetiId id)
{
	VeTinyItem * item = vetiGetItem(&instance->veti, id);
	if (item)
		item->properties.request = 1;
}

static veBool linkChargeCurrentSet(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VeDirectService * const instance = ctx;
	VeVariant linkChargeCurrent = *variant;

	if ((instance->appVersion & 0xBFFF) < 0x0118) {
		VeTinyItem * vetiItem = vetiGetItem(&instance->veti, VE_REG_LINK_CHARGE_CURRENT_PERCENTAGE);
		VeVariantValue batMaxCurrent;

		if (!veVariantIsValid(&linkChargeCurrent)) {
			/* Set to invalid... set the link charge current to 0 % */
			logI(MODULE, "/Link/ChargerCurrent is invalid");
			vetiItem->value.UN8 = 0;
		} else if (!vetiGetValue(&instance->veti, VE_REG_BAT_MAX_CURRENT, &batMaxCurrent)) {
			/* We do not have the max current (yet) so set charge current to 0 % */
			vetiItem->value.UN8 = 0;
		} else {
			float percentage;

			veVariantToFloat(&linkChargeCurrent);
			logI(MODULE, "/Link/ChargerCurrent set to %f", linkChargeCurrent.value.Float);

			percentage = 100.0 * linkChargeCurrent.value.Float / ((float) batMaxCurrent.UN16 / DIV_I);
			if (percentage < 0.0) {
				vetiItem->value.UN8 = 0;
			} else if (percentage > 100.0) {
				vetiItem->value.UN8 = 100;
			} else {
				vetiItem->value.UN8 = (un8) percentage;
			}
		}

		vetiItem->properties.valid = veTrue;
		vetiPublishItem(&instance->veti, instance->vetiHandle, VE_REG_LINK_CHARGE_CURRENT_PERCENTAGE, vetiItem);
	} else {
		/* This has absolute current settings */
		VeTinyItem * vetiItem = vetiGetItem(&instance->veti, VE_REG_LINK_CHARGE_CURRENT_LIMIT);

		if (!veVariantIsValid(&linkChargeCurrent)) {
			/* Set to invalid... set the link charge current to 0 % */
			logI(MODULE, "/Link/ChargerCurrent is invalid");
			vetiItem->value.UN16 = 0;
		} else {
			veVariantToFloat(&linkChargeCurrent);
			logI(MODULE, "/Link/ChargerCurrent set to %f", linkChargeCurrent.value.Float);

			if (linkChargeCurrent.value.Float < 0.0f) {
				vetiItem->value.UN16 = 0;
			} else if (linkChargeCurrent.value.Float > 655.34f) {
				vetiItem->value.UN16 = 0xFFFE;
			} else {
				vetiItem->value.UN16 = linkChargeCurrent.value.Float * DIV_I;
			}
		}

		vetiItem->properties.valid = veTrue;
		vetiPublishItem(&instance->veti, instance->vetiHandle, VE_REG_LINK_CHARGE_CURRENT_LIMIT, vetiItem);
	}

	return veTrue;
}

static veBool genericItemSet(struct VeItem *item, void *ctx, VeVariant *variant)
{
	VeDirectService * const instance = ctx;
	const ItemInfo * itemInfo = getItemInfoFromItem(instance, item);

	if (itemInfo) {
		const VetiId veReg = itemInfo->veReg;
		if (veReg != VE_REG_INVALID) {
			VeVariant v = *variant;
			VeTinyItem * vetiItem = vetiGetItem(&instance->veti, veReg);
			if (vetiItem == NULL) {
				logE(MODULE, "Veti item %04X could not be found. Exiting...", veReg);
				pltExit(145);
			}
			if (itemInfo->div) {
				veVariantToFloat(&v);
				v.value.Float *= itemInfo->div;
				veVariantFromFloat(&v, VE_SN32);
			} else {
				veVariantToN32(&v);
			}
			switch (vetiItem->properties.type) {
			case VETI_SN8:
				vetiItem->value.SN8 = (sn8) v.value.SN32;
				break;
			case VETI_UN8:
				vetiItem->value.UN8 = (un8) v.value.UN32;
				break;
			case VETI_SN16:
				vetiItem->value.SN16 = (sn16) v.value.SN32;
				break;
			case VETI_UN16:
				vetiItem->value.UN16 = (un16) v.value.UN32;
				break;
			case VETI_SN32:
				vetiItem->value.SN32 = v.value.SN32;
				break;
			case VETI_UN32:
				vetiItem->value.UN32 = v.value.UN32;
				break;
			case VETI_UNKNOWN:
			case VETI_PTR:
				logW(MODULE, "Unable to set veti item %04X of type %d", veReg, vetiItem->properties.type);
				return veFalse;
			}
			vetiItem->properties.valid = veTrue;
			vetiPublishItem(&instance->veti, instance->vetiHandle, veReg, vetiItem);
		}
		return veTrue;
	} else {
		char buf[256];
		veItemUid(item, buf, sizeof(buf));
		logW(MODULE, "Item %s not found in item list", buf);
		return veFalse;
	}
}

static void tick(VeDirectService *instance)
{
	if (instance->items.solar.errorSuppressTimer)
		instance->items.solar.errorSuppressTimer--;
}

static void getHistory(VeDirectService *instance)
{
	unsigned int day;

	/* We do not invalidate the veItems. In principle, they are still valid but only
	 * contain the wrong data. They will be updated when all the history items are
	 * fetched. */

	if (!vetiIsValid(&instance->veti, VE_REG_HISTORY_TOTAL) || instance->items.solar.historyDaysAvailable == 0) {
		vetiPublishRequest(&instance->veti, instance->vetiHandle, VE_REG_HISTORY_TOTAL);
	}

	for (day = 0; day < instance->items.solar.historyDaysAvailable; ++day) {
		vetiPublishRequest(&instance->veti, instance->vetiHandle, VE_REG_HISTORY_DAY00 + day);
	}
}

static void setDailyItemValue(VeDirectService *instance, const ItemInfo * item, unsigned int day, sn32 value, veBool valid)
{
	VeItem * i = (VeItem *)(((SolarHistDay *) instance_ptr(instance, item->item_offset)) + day);
	if (valid) {
		VeVariant v;
		itemsSetItem(i, veVariantSn32(&v, value), item);
	} else {
		veItemInvalidate(i);
	}
}

static void setDailyItemBuffer(VeDirectService *instance, const ItemInfo * item, unsigned int day, const MpptDayHistory * histDaily)
{
	VeVariant v;
	itemsSetBufferPtr((VeItem *)(((SolarHistDay *) instance_ptr(instance, item->item_offset)) + day),
					   veVariantBuf(&v, histDaily, sizeof(*histDaily)), &instance->items.solar.historyDaily[day], sizeof(*histDaily));
}

static void setTotalItemValue(VeDirectService *instance, const ItemInfo * item, sn32 value, veBool valid)
{
	VeItem * i = instance_ptr(instance, item->item_offset);
	if (valid) {
		VeVariant v;
		itemsSetItem(i, veVariantSn32(&v, value), item);
	} else {
		veItemInvalidate(i);
	}
}

static void setTotalItemBuffer(VeDirectService *instance, const ItemInfo * item, const MpptTotalsHistory * histTotal)
{
	VeVariant v;
	itemsSetBufferPtr(instance_ptr(instance, item->item_offset),
					   veVariantBuf(&v, histTotal, sizeof(*histTotal)), &instance->items.solar.historyTotal, sizeof(*histTotal));
}

static un32 determineCapabilities (VeDirectService * instance)
{
	un32 value;
	un16 pid = instance->prodId;
	un16 appVer = instance->appVersion;

	switch (pid) {
	case 0x0300:
		if ((appVer & 0xF000) && ((appVer & 0x00F0) >= 0x00A0)) {
			value = 0x0001;
			logE(MODULE, "alternative release candidate numbering for MPPT model %04X with firmware %04X. Using capabilities %08X", pid, appVer, value);
		} else {
			appVer = appVer & 0x0FFF;
			if (appVer <= 0x0107) {
				value = 0x0019;
			} else if (appVer <= 0x0112) {
				value = 0x0009;
			} else if (appVer <= 0x0114) {
				value = 0x0001;
			} else if (appVer <= 0x0115) {
				value = 0x1001;
			} else {
				value = 0x1001;
				logE(MODULE, "Unknown capabilities for MPPT model %04X with firmware %04X. Using capabilities %08X", pid, appVer, value);
			}
		}
		break;
	case 0xA042:
	case 0xA043:
	case 0xA04C:
		if ((appVer & 0xF000) && ((appVer & 0x00F0) >= 0x00A0)) {
			value = 0x0005;
			logE(MODULE, "alternative release candidate numbering for MPPT model %04X with firmware %04X. Using capabilities %08X", pid, appVer, value);
		} else {
			appVer = appVer & 0x0FFF;
			if (appVer <= 0x0107) {
				value = 0x001D;
			} else if (appVer <= 0x0112) {
				value = 0x000D;
			} else if (appVer <= 0x0114) {
				value = 0x0005;
			} else if (appVer <= 0x0115) {
				value = 0x1E05;
			} else {
				value = 0x1E05;
				logE(MODULE, "Unknown capabilities for MPPT model %04X with firmware %04X. Using capabilities %08X", pid, appVer, value);
			}
		}
		break;
	case 0xA040:
	case 0xA041:
	case 0xA044:
	case 0xA045:
	case 0xA046:
	case 0xA047:
	case 0xA048:
	case 0xA049:
	case 0xA04A:
	case 0xA04B:
		if ((appVer & 0xF000) && ((appVer & 0x00F0) >= 0x00A0)) {
			value = 0x2006;
			logE(MODULE, "alternative release candidate numbering for MPPT model %04X with firmware %04X. Using capabilities %08X", pid, appVer, value);
		} else {
			appVer = appVer & 0x0FFF;
			if (appVer <= 0x0107) {
				value = 0x201E;
			} else if (appVer <= 0x0112) {
				value = 0x200E;
			} else if (appVer <= 0x0114) {
				value = 0x2106;
			} else if (appVer <= 0x0115) {
				value = 0x2506;
			} else {
				value = 0x2506;
				logE(MODULE, "Unknown capabilities for MPPT model %04X with firmware %04X. Using capabilities %08X", pid, appVer, value);
			}
		}
		break;
	default:
		value = 0;
		logE(MODULE, "Unknown capabilities for MPPT model %04X with firmware %04X. Using capabilities %08X", pid, appVer, value);
	}

	logI(MODULE, "Determined VE_REG_CAPABILITIES1 for MPPT model %04X with firmware %04X. Using %08X", pid, appVer, value);
	return value;
}

static void updateFromCapabilities (VeDirectService *instance)
{
	const ItemInfo * itemInfo;
	VeItem * item;
	VeVariant v;
	un32 cap1 = veItemLocalValue(&instance->items.solar.capabilities.capabilities1, &v)->value.UN32;

	itemInfo = getItemInfoFromId(VE_REG_LINK_NETWORK_MODE);
	item = instance_ptr(instance, itemInfo->item_offset);
	if ((cap1 & VE_REG_CAPABILITIES1_HAS_PARALLEL_SUPPORT) && !veItemParent(item)) {
		itemsAddItemByInfo(&instance->root, item, itemInfo, genericItemSet, instance);
		vetiPublishRequest(&instance->veti, instance->vetiHandle, itemInfo->veReg);
	}

	itemInfo = getItemInfoFromId(VE_REG_LINK_NETWORK_STATUS);
	item = instance_ptr(instance, itemInfo->item_offset);
	if ((cap1 & VE_REG_CAPABILITIES1_HAS_PARALLEL_SUPPORT) && !veItemParent(item)) {
		itemsAddItemByInfo(&instance->root, item, itemInfo, genericItemSet, instance);
		vetiPublishRequest(&instance->veti, instance->vetiHandle, itemInfo->veReg);
	}

	itemInfo = getItemInfoFromId(VE_REG_LINK_VSET);
	item = instance_ptr(instance, itemInfo->item_offset);
	if ((cap1 & VE_REG_CAPABILITIES1_HAS_HUB1_SUPPORT) && !veItemParent(item)) {
		itemsAddItemByInfo(&instance->root, item, itemInfo, genericItemSet, instance);
		vetiPublishRequest(&instance->veti, instance->vetiHandle, itemInfo->veReg);
	}

	if ((instance->appVersion & 0xBFFF) < 0x0118) {
		itemInfo = getItemInfoFromId(VE_REG_LINK_CHARGE_CURRENT_PERCENTAGE);
		item = instance_ptr(instance, itemInfo->item_offset);
	} else {
		itemInfo = getItemInfoFromId(VE_REG_LINK_CHARGE_CURRENT_LIMIT);
		item = instance_ptr(instance, itemInfo->item_offset);
	}
	if ((cap1 & VE_REG_CAPABILITIES1_HAS_BMS_SUPPORT) && !veItemParent(item)) {
		itemsAddItemByInfo(&instance->root, item, itemInfo, linkChargeCurrentSet, instance);
		vetiPublishRequest(&instance->veti, instance->vetiHandle, itemInfo->veReg);
	}

	itemInfo = getItemInfoFromId(VE_REG_BAT_BMS_PRESENT);
	item = instance_ptr(instance, itemInfo->item_offset);
	if ((cap1 & VE_REG_CAPABILITIES1_HAS_BMS_SUPPORT) && !veItemParent(item)) {
		itemsAddItemByInfo(&instance->root, item, itemInfo, genericItemSet, instance);
		vetiPublishRequest(&instance->veti, instance->vetiHandle, itemInfo->veReg);
	}

	itemInfo = getItemInfoFromId(VE_REG_DC_OUTPUT_STATUS);
	item = instance_ptr(instance, itemInfo->item_offset);
	if ((cap1 & (VE_REG_CAPABILITIES1_HAS_LOAD_OUTPUT | VE_REG_CAPABILITIES1_HAS_VIRTUAL_LOAD_OUTPUT)) && !veItemParent(item)) {
		itemsAddItemByInfo(&instance->root, item, itemInfo, genericItemSet, instance);
		vetiPublishRequest(&instance->veti, instance->vetiHandle, itemInfo->veReg);
	}

	itemInfo = getItemInfoFromId(VE_REG_DC_OUTPUT_CURRENT);
	item = instance_ptr(instance, itemInfo->item_offset);
	if ((cap1 & VE_REG_CAPABILITIES1_HAS_LOAD_OUTPUT) && !veItemParent(item)) {
		itemsAddItemByInfo(&instance->root, item, itemInfo, genericItemSet, instance);
		vetiPublishRequest(&instance->veti, instance->vetiHandle, itemInfo->veReg);
	}

	itemInfo = getItemInfoFromId(VE_REG_RELAY_CONTROL);
	item = instance_ptr(instance, itemInfo->item_offset);
	if ((cap1 & (VE_REG_CAPABILITIES1_HAS_VIRTUAL_RELAY | VE_REG_CAPABILITIES1_HAS_ALARM_RELAY1))
			&& !veItemParent(item)) {
		itemsAddItemByInfo(&instance->root, item, itemInfo, genericItemSet, instance);
		vetiPublishRequest(&instance->veti, instance->vetiHandle, itemInfo->veReg);
	}
}

static void itemUpdate (VeDirectService *instance, const VetiId id, VeVariantValue value, VetiTypes type, const veBool valid)
{
	const ItemInfo * itemInfo = NULL;
	un16 oldDay0SequenceNr = instance->items.solar.day0SequenceNr;
	un8 oldHistoryDaysAvailable = instance->items.solar.historyDaysAvailable;
	veBool wasHistory = veFalse;
	VeVariant v;

	logI(MODULE, "itemUpdate for id %04X -> %08X (%s)", id, value.UN32, valid ? "valid" : "invalid");

	/* First handle the special cases */
	if (id == VE_REG_HSDS) {
		if (valid && value.UN16 != VE_INVALID_UN16) {
			instance->items.solar.day0SequenceNr = value.UN16;
		}
		wasHistory = veTrue;
	}

	if (id == VE_REG_HISTORY_TOTAL) {
		if (valid && value.CPtr) {
			const MpptTotalsHistory * histTotal = value.CPtr;
			unsigned int error;

			setTotalItemValue(instance, &historyTotalItemInfo[totalYieldSystem], histTotal->systemYield, validUN32(histTotal->systemYield));
			setTotalItemValue(instance, &historyTotalItemInfo[totalYieldUser], histTotal->userYield, validUN32(histTotal->userYield));
			setTotalItemValue(instance, &historyTotalItemInfo[totalMaxPvVoltage], histTotal->maxPanelVoltage, validUN16(histTotal->maxPanelVoltage));
			setTotalItemValue(instance, &historyTotalItemInfo[totalMaxBatteryVoltage], histTotal->maxBatteryVoltage, validUN16(histTotal->maxBatteryVoltage));
			setTotalItemValue(instance, &historyTotalItemInfo[totalMinBatteryVoltage], histTotal->minBatteryVoltage, validUN16(histTotal->minBatteryVoltage));
			for (error = 0; error < 4; ++error)
				setTotalItemValue(instance, &historyTotalItemInfo[totalLastError0 + error], histTotal->errors.chargerErrors[error], veTrue);
			setTotalItemValue(instance, &historyTotalItemInfo[totalDaysAvailable], histTotal->daysAvailable + 1, veTrue);

			setTotalItemBuffer(instance, &historyTotalItemInfo[totalRaw], histTotal);

			instance->items.solar.historyDaysAvailable = histTotal->daysAvailable + 1;
		}
		wasHistory = veTrue;
	}

	if (id >= VE_REG_HISTORY_DAY00 && id <= VE_REG_HISTORY_DAY30) {
		if (valid && value.CPtr) {
			const MpptDayHistory * histDaily = value.CPtr;
			const unsigned int day = id - VE_REG_HISTORY_DAY00;
			unsigned int error;

			setDailyItemValue(instance, &historyDailyItemInfo[dailyAbsTime], day, histDaily->timeInAbsorption, validUN16(histDaily->timeInAbsorption));
			setDailyItemValue(instance, &historyDailyItemInfo[dailyBulkTime], day, histDaily->timeInBulk, validUN16(histDaily->timeInBulk));
			setDailyItemValue(instance, &historyDailyItemInfo[dailyConsumption], day, histDaily->consumed, validUN32(histDaily->consumed));
			setDailyItemValue(instance, &historyDailyItemInfo[dailyFloatTime], day, histDaily->timeInFloat, validUN16(histDaily->timeInFloat));
			for (error = 0; error < 4; ++error)
				setDailyItemValue(instance, &historyDailyItemInfo[dailyLastError0 + error], day, histDaily->errors.chargerErrors[error], veTrue);
			setDailyItemValue(instance, &historyDailyItemInfo[dailyMaxBatteryVoltage], day, histDaily->maxBatteryVoltage, validUN16(histDaily->maxBatteryVoltage));
			setDailyItemValue(instance, &historyDailyItemInfo[dailyMaxCurrent], day, histDaily->maxBatteryCurrent, validUN16(histDaily->maxBatteryCurrent));
			setDailyItemValue(instance, &historyDailyItemInfo[dailyMaxPower], day, histDaily->maxPower, validUN32(histDaily->maxPower));
			setDailyItemValue(instance, &historyDailyItemInfo[dailyMaxPvVoltage], day, histDaily->maxPanelVoltage, validUN16(histDaily->maxPanelVoltage));
			setDailyItemValue(instance, &historyDailyItemInfo[dailyMinBatteryVoltage], day, histDaily->minBatteryVoltage, validUN16(histDaily->minBatteryVoltage));
			setDailyItemValue(instance, &historyDailyItemInfo[dailyYield], day, histDaily->yield, validUN32(histDaily->yield));
			setDailyItemValue(instance, &historyDailyItemInfo[dailySeqNr], day, histDaily->daySequenceNumber, veTrue);

			setDailyItemBuffer(instance, &historyDailyItemInfo[dailyRaw], day, histDaily);

			if (id == VE_REG_HISTORY_DAY00) {
				instance->items.solar.day0SequenceNr = histDaily->daySequenceNumber;
			}
		}
		wasHistory = veTrue;
	}

	if (wasHistory) {
		if (oldDay0SequenceNr != instance->items.solar.day0SequenceNr
				|| oldHistoryDaysAvailable != instance->items.solar.historyDaysAvailable) {
			getHistory(instance);
		}

		return;
	}

	/* Values that need changes before normal processing */
#ifdef ITEM_TREE
	if (id == VE_REG_CAPABILITIES1) {
		if (!valid) {
			value.UN32 = determineCapabilities(instance);
		}
	}
#endif

	if (id == VE_REG_LINK_CHARGE_CURRENT_PERCENTAGE) {
		/* Convert from percentage to absolute current */
		VeVariantValue batMaxCurrent;

		if (valid
				&& value.UN8 != 0xFF
				&& vetiGetValue(&instance->veti, VE_REG_BAT_MAX_CURRENT, &batMaxCurrent)
				&& batMaxCurrent.UN16 != 0xFFFF) {
			value.UN16 = (un16) ((un32) batMaxCurrent.UN16 * (un32) value.UN8 / 100);
		} else {
			value.UN16 = 0xFFFF;
		}
		type = VETI_UN16;
	}

	if (id == VE_REG_CHR_ERROR_CODE) {
		switch (value.UN8)
		{
		/* Always suppress error 19 (current polarity reversed) */
		case CHARGER_ERROR_CHARGER_CURRENT_REVERSED:
			instance->items.solar.errorSuppressTimer = 0xffff;
			break;
			/* Suppress error 21 for 5min (current sensor issue) */
		case CHARGER_ERROR_CHARGER_CURRENT_SENSE:
			/* Trigger the timer on-change */
			if (instance->items.solar.errorCopy != value.UN8)
				instance->items.solar.errorSuppressTimer = 300 /* seconds */ * TICK_FREQ;
			break;
		default:
			instance->items.solar.errorSuppressTimer = 0;
		}

		/* Save copy for comparison */
		instance->items.solar.errorCopy = value.UN8;

		/* Suppress */
		if (instance->items.solar.errorSuppressTimer)
			value.UN8 = CHARGER_ERROR_NONE;
	}

	if (id == VE_REG_DEVICE_STATE) {
		/* Check for suppressed errors, then set state from fault to off. */
		if (value.UN8 == N2K_CONVERTER_STATE_FAULT && instance->items.solar.errorSuppressTimer) {
			value.UN8 = N2K_CONVERTER_STATE_OFF;
		}
	}

	/* Normal processing */
	/* Lets find the ItemInfo */
	itemInfo = getItemInfoFromId(id);

	if (!itemInfo) {
		/* Not much we can do */
	} else if (valid) {
		switch (type) {
		case VETI_UN8:
			veVariantToN32(veVariantUn8(&v, value.UN8));
			break;
		case VETI_SN8:
			veVariantToN32(veVariantSn8(&v, value.SN8));
			break;
		case VETI_UN16:
			veVariantToN32(veVariantUn16(&v, value.UN16));
			break;
		case VETI_SN16:
			veVariantToN32(veVariantSn16(&v, value.SN16));
			break;
		case VETI_UN32:
			veVariantToN32(veVariantUn32(&v, value.UN32));
			break;
		case VETI_SN32:
			veVariantToN32(veVariantSn32(&v, value.SN32));
			break;
		/* We don't know the size of the item, so really nothing we can do but handle
		 * it above with the specific items. */
		case VETI_PTR:
			logE(MODULE, "Received VETI_PTR item update for %04X. We can not handle this in a generic way. Exiting...", id);
			pltExit(146);

		case VETI_UNKNOWN:
			logE(MODULE, "Received VETI_UKNOWN item update for %04X. We can not handle this in a generic way. Exiting...", id);
			pltExit(147);
		}

		itemsSetItem(instance_ptr(instance, itemInfo->item_offset), &v, itemInfo);
	} else {
		veItemInvalidate(instance_ptr(instance, itemInfo->item_offset));
	}

	if (id == VE_REG_CHR_ERROR_CODE) {
		/* It could be that this one is set after state changed to fault, so also recalculate state after the error changed */
		VeVariantValue deviceState;
		vetiGetValue(&instance->veti, VE_REG_DEVICE_STATE, &deviceState);
		if (deviceState.UN8 == N2K_CONVERTER_STATE_FAULT && instance->items.solar.errorSuppressTimer) {
			veVariantToN32(veVariantUn8(&v, N2K_CONVERTER_STATE_OFF));
		} else {
			veVariantToN32(veVariantUn8(&v, deviceState.UN8));
		}
		veItemOwnerSet(&instance->items.solar.state, &v);
	}

	/* The PV current is not available from text frames, so for regular updates we would need to fetch
	 * using hex, disrupting text traffic. Solution is to always calculate the current from PV power and
	 * PV voltage, which are available in text frames. */
	if (id == VE_REG_DC_INPUT_VOLTAGE || id == VE_REG_DC_INPUT_POWER) {
		if (veItemIsValid(&instance->items.solar.dcStatusPvPanel.power)
				&& veItemIsValid(&instance->items.solar.dcStatusPvPanel.voltage)) {
			VeVariant powerVar;
			float power = veItemLocalValue(&instance->items.solar.dcStatusPvPanel.power, &powerVar)->value.Float;
			VeVariant voltageVar;
			float voltage = veItemLocalValue(&instance->items.solar.dcStatusPvPanel.voltage, &voltageVar)->value.Float;
			VeVariant currentVar;
			float current = 0.0;
			if (voltage > 0.0)
				current = power / voltage;
			veItemOwnerSet(&instance->items.solar.dcStatusPvPanel.current, veVariantFloat(&currentVar, current));
		}
	}
#ifdef ITEM_TREE
	if (id == VE_REG_BAT_MAX_CURRENT) {
		/* The Maximum battery current changed, so we need to recalculate the charge percentage */
		linkChargeCurrentSet(&instance->items.solar.link.chargeCurrentSet, instance, veItemLocalValue(&instance->items.solar.link.chargeCurrentSet, &v));
	}

	if (id == VE_REG_CAPABILITIES1) {
		updateFromCapabilities(instance);
	}

	if (id == VE_REG_LINK_NETWORK_STATUS) {
		/* VE_REG_LINK_NETWORK_MODE is not sent on change, so request when NETWORK_STATUS is received,
		 * because network status is sent on change */
		vetiPublishRequest(&instance->veti, instance->vetiHandle, VE_REG_LINK_NETWORK_MODE);
	}
	if (id == VE_REG_BAT_BMS_PRESENT) {
		vetiPublishRequest(&instance->veti, instance->vetiHandle, VE_REG_LINK_NETWORK_MODE);
	}
#endif
}

static void connectedEvent(VeDirectService *instance, un16 pid)
{
	un16 i;
	unsigned int day;

	logI(MODULE, "found %s", veProductGetName(pid));

	memset(&instance->items.solar, 0, sizeof(instance->items.solar));

	instance->items.solar.day0SequenceNr = VE_INVALID_UN16;
	instance->items.solar.historyDaysAvailable = 0;

	/* Set the request property of the items, so they are always requested and we do not get the
	 * cached value from veti. */
	setVetiRequest(instance, VE_REG_LINK_NETWORK_MODE);
	setVetiRequest(instance, VE_REG_HISTORY_TOTAL);
	for (i = VE_REG_HISTORY_DAY00; i <= VE_REG_HISTORY_DAY30; ++i) {
		setVetiRequest(instance, i);
	}

	/* Build items */
	for (i = 0; i < ARRAY_LENGTH(itemInfo); i++) {
		if (!itemInfo[i].conditional) {
			VeItem * item = instance_ptr(instance, itemInfo[i].item_offset);
			itemsAddItemByInfo(&instance->root, item, &itemInfo[i], genericItemSet, instance);
		}
	}

	for (i = 0; i < ARRAY_LENGTH(historyTotalItemInfo); ++i) {
		VeItem * item = instance_ptr(instance, historyTotalItemInfo[i].item_offset);
		itemsAddItemByInfo(&instance->root, item, &historyTotalItemInfo[i], NULL, NULL);
	}

	for (day = 0; day < ARRAY_LENGTH(instance->items.solar.histDay); ++day) {
		for (i = 0; i < ARRAY_LENGTH(historyDailyItemInfo); ++i) {
			char buf[256];
			ItemInfo item = historyDailyItemInfo[i];
			VeItem * ve_item = (VeItem *)(((SolarHistDay *)instance_ptr(instance, item.item_offset)) + day);
			ve_snprintf(buf, sizeof(buf), item.id, day);
			item.id = buf;
			itemsAddItemByInfo(&instance->root, ve_item, &item, NULL, NULL);
		}
	}

	/* Request item values */
	for (i = 0; i < ARRAY_LENGTH(requestItems); i++) {
		vetiPublishRequest(&instance->veti, instance->vetiHandle, requestItems[i]);
	}

	getHistory(instance);
}

static veBool hasSerialNumber(un16 pid)
{
	VE_UNUSED(pid);

	return veTrue;
}

static const VeDirectProduct veDirectSolar = {
	NULL,
	tick,
	NULL,
	itemUpdate,
	hasSerialNumber,

	connectedEvent,
	NULL,

#ifdef ITEM_TREE
	solarName,
#endif
};

const VeDirectProduct *veDirectSolarFunc(un16 pid)
{
	if (pid == VE_PROD_ID_BLUE_SOLAR_MPPT_70_15)
		return &veDirectSolar;

	if ((pid & ~0x3F) == 0xA040)
		return &veDirectSolar;

	return NULL;
}
